import logging
from copy import deepcopy

import numpy as np
import pytest
import xarray as xr
import xclim as xc

try:
    import xesmf as xe
except ImportError:
    xe = None

from xclim.testing.helpers import test_timeseries as timeseries
from xclim.testing.utils import nimbus

import xscen as xs

LOGGER = logging.getLogger(__name__)


class TestEnsembleStats:
    @staticmethod
    def make_ensemble(n):
        ens = []
        for i in range(n):
            tmp = timeseries(
                np.arange(1, 5) + i / 10,
                variable="tas",
                start="1981-01-01",
                freq="30YS",
                as_dataset=True,
            )
            tmp = tmp.assign_coords(
                {
                    "horizon": xr.DataArray(
                        ["1981-2010", "2011-2040", "2041-2070", "2071-2100"],
                        dims="time",
                    )
                }
            )
            tmp = tmp.rename({"tas": "tg_mean"})

            ens.append(tmp)

        return ens

    def test_input_type(self, tmpdir):
        ens_dict = self.make_ensemble(10)
        out_dict = xs.ensemble_stats(
            ens_dict, statistics={"ensemble_mean_std_max_min": None}
        )
        ens_ds = xr.concat(ens_dict, dim="realization")
        out_ds = xs.ensemble_stats(
            ens_ds, statistics={"ensemble_mean_std_max_min": None}
        )
        paths = []
        for i, ds in enumerate(ens_dict):
            paths.append(tmpdir / f"ens{i}.zarr")
            ds.to_zarr(paths[-1])
        out_zarr = xs.ensemble_stats(
            paths, statistics={"ensemble_mean_std_max_min": None}
        ).compute()

        assert out_dict.equals(out_ds)
        # The zarr introduced some rounding errors
        assert np.round(out_zarr, 10).equals(np.round(out_dict, 10))

    @pytest.mark.parametrize(
        "weights, to_level", [(None, None), (np.arange(1, 11), "for_tests")]
    )
    def test_weights(self, weights, to_level):
        ens = self.make_ensemble(10)
        if weights is not None:
            weights = xr.DataArray(
                weights, dims="realization", coords={"realization": np.arange(0, 10)}
            )

        if to_level is None:
            out = xs.ensemble_stats(
                ens,
                statistics={
                    "ensemble_mean_std_max_min": None,
                    "ensemble_percentiles": {"split": False},
                },
                weights=weights,
            )
        else:
            out = xs.ensemble_stats(
                ens,
                statistics={
                    "ensemble_mean_std_max_min": None,
                    "ensemble_percentiles": {"split": False},
                },
                weights=weights,
                to_level=to_level,
            )

        # Check that the output has the right attributes
        assert out.attrs["ensemble_size"] == 10
        assert (
            out.attrs["cat:processing_level"] == to_level
            if to_level is not None
            else "ensemble"
        )
        assert "realization" not in out.dims
        # Check that the output has the right variables and results
        np.testing.assert_array_equal(
            out.tg_mean_mean,
            [1.45, 2.45, 3.45, 4.45] if weights is None else [1.6, 2.6, 3.6, 4.6],
        )
        np.testing.assert_array_almost_equal(
            out.tg_mean_stdev,
            [0.2872] * 4 if weights is None else [0.2449] * 4,
            decimal=4,
        )
        np.testing.assert_array_almost_equal(
            out.tg_mean.sel(percentiles=90),
            [1.81, 2.81, 3.81, 4.81] if weights is None else [1.87, 2.87, 3.87, 4.87],
            decimal=2,
        )

    @pytest.mark.parametrize("fractions", ["only", "both", "nested", "missing"])
    def test_robustness_input(self, fractions):
        ens = self.make_ensemble(10)

        weights = None
        if fractions == "only":
            statistics = {
                "robustness_fractions": {"test": "threshold", "abs_thresh": 2.5}
            }
        elif fractions == "both":
            statistics = {
                "robustness_fractions": {"test": "threshold", "abs_thresh": 2.5},
                "robustness_categories": {
                    "categories": ["robust", "non-robust"],
                    "thresholds": [(0.5, 0.5), (0.5, 0.5)],
                    "ops": [(">=", ">="), ("<", "<")],
                },
            }
        elif fractions == "nested":
            weights = xr.DataArray(
                [0] * 5 + [1] * 5, dims="realization"
            )  # Mostly to check that this is put in the nested dict
            statistics = {
                "robustness_categories": {
                    "robustness_fractions": {"test": "threshold", "abs_thresh": 2.5},
                    "categories": ["robust", "non-robust"],
                    "thresholds": [(0.5, 0.5), (0.5, 0.5)],
                    "ops": [(">=", ">="), ("<", "<")],
                }
            }
        elif fractions == "missing":
            statistics = {
                "robustness_categories": {
                    "categories": ["robust", "non-robust"],
                    "thresholds": [(0.5, 0.5), (0.5, 0.5)],
                    "ops": [(">=", ">="), ("<", "<")],
                }
            }
            with pytest.raises(
                ValueError,
                match="'robustness_categories' requires 'robustness_fractions'",
            ):
                xs.ensemble_stats(ens, statistics=statistics)
            return

        out = xs.ensemble_stats(ens, statistics=statistics, weights=weights)

        # Output length should be 1 if nested, >=5 otherwise
        if fractions == "nested":
            assert len(out.data_vars) == 1
        else:
            assert len(out.data_vars) >= 5
        if fractions in ["only", "both"]:
            np.testing.assert_array_equal(out.tg_mean_changed, [0, 0.4, 1, 1])
            np.testing.assert_array_equal(out.tg_mean_agree, [1, 1, 1, 1])
        if fractions in ["both", "nested"]:
            np.testing.assert_array_equal(
                out.tg_mean_robustness_categories,
                {"both": [99, 99, 1, 1], "nested": [99, 1, 1, 1]}[fractions],
            )
            np.testing.assert_array_equal(
                out.tg_mean_robustness_categories.attrs["flag_descriptions"],
                ["robust", "non-robust"],
            )

    @pytest.mark.parametrize("symbol", ["rel.", "relative", "*", "/", "pct.", "abs."])
    def test_robustness_reldelta(self, caplog, symbol):
        ens = self.make_ensemble(10)
        for e in ens:
            e["tg_mean"].attrs["delta_kind"] = symbol

        with caplog.at_level(logging.INFO):
            xs.ensemble_stats(
                ens,
                statistics={
                    "robustness_fractions": {"test": "threshold", "abs_thresh": 2.5}
                },
            )
        if symbol in ["rel.", "relative", "*", "/"]:
            assert "Relative delta detected" in caplog.text
        else:
            assert "Relative delta detected" not in caplog.text

    @pytest.mark.parametrize("common_attrs_only", [True, False])
    def test_common_attrs_only(self, common_attrs_only):
        ens = self.make_ensemble(10)
        for e in range(len(ens)):
            ens[e].attrs["foo"] = "bar"
            ens[e].attrs[f"bar{e}"] = "foo"

        out = xs.ensemble_stats(
            ens,
            statistics={"ensemble_mean_std_max_min": None},
            common_attrs_only=common_attrs_only,
        )
        assert out.attrs.get("foo", None) == "bar"
        assert ("bar0" not in out.attrs) == common_attrs_only

    def test_errors(self):
        ens = self.make_ensemble(10)

        # Warning if the statistic does not support weighting
        weights = xr.DataArray([0] * 5 + [1] * 5, dims="realization")
        with pytest.warns(UserWarning, match="Weighting is not supported"):
            with pytest.raises(
                TypeError
            ):  # kkz is not actually supported here, but it's one of the few that will not support weighting
                xs.ensemble_stats(
                    ens, statistics={"kkz_reduce_ensemble": None}, weights=weights
                )

        # Error if you try to use a relative delta with a reference dataset
        for e in ens:
            e["tg_mean"].attrs["delta_kind"] = "rel."
        ref = weights
        with pytest.raises(
            ValueError, match="is a delta, but 'ref' was still specified."
        ):
            xs.ensemble_stats(
                ens,
                statistics={
                    "robustness_fractions": {
                        "test": "threshold",
                        "abs_thresh": 2.5,
                        "ref": ref,
                    }
                },
            )


class TestGenerateWeights:
    @staticmethod
    def make_ensemble():
        out = {}

        gcms = {
            "CCCma": ["CanESM2"],
            "GFDL": ["GFDL-ESM2M", "GFDL-ESM2G"],
            "CSIRO-QCCCE": ["CSIRO-Mk3-6-0"],
            "ECMWF": ["EC-EARTH"],
        }
        n_members = {
            "CanESM2": 3,
            "GFDL-ESM2M": 1,
            "GFDL-ESM2G": 2,
            "CSIRO-Mk3-6-0": 5,
            "EC-EARTH": 1,
        }
        rcms = {
            "CRCM5": ["CanESM2-r1i1p1"],
            "CanRCM4": ["CanESM2-r1i1p1", "CanESM2-r3i1p1"],
            "HIRHAM5": ["GFDL-ESM2M-r1i1p1", "EC-EARTH-r1i1p1"],
            "RegCM4": ["EC-EARTH-r1i1p1"],
        }

        for institution in ["CCCma", "GFDL", "CSIRO-QCCCE", "ECMWF"]:
            for gcm in gcms[institution]:
                members = [f"r{i}i1p1" for i in range(1, n_members[gcm] + 1)]
                for member in members:
                    has_gcm = (
                        []
                        if (
                            (gcm == "CanESM2" and member == "r3i1p1")
                            or gcm == "EC-EARTH"
                        )
                        else [gcm]
                    )
                    rcm = [r for r in rcms if f"{gcm}-{member}" in rcms[r]]
                    for model in has_gcm + rcm:
                        ds = timeseries(
                            np.arange(1, 5),
                            variable="tas",
                            start="1981-01-01",
                            freq="30YS",
                            as_dataset=True,
                        )
                        ds["horizon"] = xr.DataArray(
                            ["1981-2010", "2041-2070", "+2C", "+4C"], dims="time"
                        )
                        ds = ds.swap_dims({"time": "horizon"}).drop_vars("time")
                        if f"{gcm}-{member}" == "CanESM2-r3i1p1":
                            ds = ds.where(ds.horizon == "1981-2010")
                        elif gcm == "GFDL-ESM2G":
                            ds = ds.where(~ds.horizon.str.startswith("+"))
                        elif gcm in ["EC-EARTH", "CSIRO-Mk3-6-0"]:
                            ds = ds.where(ds.horizon != "+4C")

                        ds.attrs = {
                            "cat:institution": institution,
                            "cat:member": member,
                            "cat:experiment": "rcp85" if gcm == "EC-EARTH" else "rcp45",
                        }
                        if model == gcm:
                            ds.attrs["cat:source"] = model
                            ds.attrs["cat:activity"] = "CMIP5"
                        else:
                            ds.attrs["cat:source"] = model
                            ds.attrs["cat:driving_model"] = gcm
                            ds.attrs["cat:activity"] = "CORDEX"

                        out[
                            f"{institution}-{gcm}-{ds.attrs['cat:experiment']}-{member}-{model}"
                        ] = ds

                        if gcm == "CanESM2":
                            ds2 = ds.copy()
                            ds2.attrs["cat:experiment"] = "rcp85"
                            out[
                                f"{institution}-{gcm}-{ds2.attrs['cat:experiment']}-{member}-{model}"
                            ] = ds2

                            if model == "CRCM5" and member == "r1i1p1":
                                for i in range(10):
                                    ds2 = ds.copy()
                                    ds2.attrs["cat:experiment"] = "rcp85"
                                    ds2.attrs["cat:member"] = f"xr{i}r1i1p1"
                                    ds2.attrs["cat:activity"] = "ClimEx"
                                    out[
                                        f"{institution}-{gcm}-{ds2.attrs['cat:experiment']}-{ds2.attrs['cat:member']}-{model}"
                                    ] = ds2
                        elif gcm == "CSIRO-Mk3-6-0" and member == "r1i1p1":
                            ds2 = ds.copy()
                            ds2.attrs["cat:experiment"] = "rcp85"
                            ds2.attrs["cat:source"] = "CSIRO2"
                            out[
                                f"{institution}-CSIRO2-{member}-{ds2.attrs['cat:experiment']}-CSIRO2"
                            ] = ds2
        return {o: out[o] for o in sorted(out)}

    @staticmethod
    def make_ensemble_rcm(ens):
        ens_rcm = ens.copy()
        for k in ens.keys():
            if "cat:driving_model" not in ens_rcm[k].attrs.keys():
                ens_rcm.pop(k)
        return ens_rcm

    ens = make_ensemble.__func__()
    ens_rcm = make_ensemble_rcm.__func__(ens)

    @staticmethod
    def make_answer(independence_level, exp_weights, skipna):
        answers = {
            "model-True": np.concatenate(
                (
                    np.array(
                        [
                            [0.14285, 0.14285, 0.16666, 0.2],
                            [0.07142, 0.07142, 0.08333, 0.1],
                            [0.07142, 0.14285, 0.16666, 0.2],
                            [0.07142, 0.07142, 0.08333, 0.1],
                            [0.07142, 0, 0, 0],
                        ]
                    ),  # CanESM2 family, RCP4.5
                    np.array(
                        [
                            [0.01515, 0.01515, 0.01515, 0.03030],
                            [0.08333, 0.08333, 0.08333, 0.16666],
                            [0.08333, 0.16666, 0.16666, 0.33333],
                            [0.08333, 0.08333, 0.08333, 0.16666],
                            [0.08333, 0, 0, 0],
                        ]
                    ),  # CanESM2 family, RCP8.5
                    np.repeat(
                        np.array([[0.01515, 0.01515, 0.01515, 0.03030]]), 10, 0
                    ),  # ClimEx
                    np.repeat(
                        np.array([[0.02857, 0.02857, 0.03333, 0]]), 5, 0
                    ),  # CSIRO-Mk6, RCP4.5
                    np.array([[0.16666, 0.16666, 0.16666, 0]]),  # CSIRO2, RCP8.5
                    np.array(
                        [[0.16666, 0.16666, 0.16666, 0], [0.16666, 0.16666, 0.16666, 0]]
                    ),  # EC-EARTH RCMs, RCP8.5
                    np.array(
                        [[0.07142, 0.07142, 0, 0], [0.07142, 0.07142, 0, 0]]
                    ),  # GFDL-ESM2G family, RCP4.5
                    np.array(
                        [
                            [0.14285, 0.14285, 0.16666, 0.2],
                            [0.14285, 0.14285, 0.16666, 0.2],
                        ]
                    ),  # GFDL-ESM2M family, RCP4.5
                )
            ),
            "model-False": np.concatenate(
                (
                    np.array(
                        [
                            [0.08333] * 4,
                            [0.25] * 4,
                            [0.25] * 1 + [0.5] * 3,
                            [0.25] * 4,
                            [0.25] * 1 + [0] * 3,
                        ]
                    ),  # CanESM2 family, RCP4.5
                    np.array(
                        [
                            [0.08333] * 4,
                            [0.25] * 4,
                            [0.25] * 1 + [0.5] * 3,
                            [0.25] * 4,
                            [0.25] * 1 + [0] * 3,
                        ]
                    ),  # CanESM2 family, RCP8.5
                    np.repeat(np.array([[0.08333] * 4]), 10, 0),  # ClimEx
                    np.repeat(
                        np.array([[0.2] * 3 + [0] * 1]), 5, 0
                    ),  # CSIRO-Mk6, RCP4.5
                    np.array([[1] * 3 + [0] * 1]),  # CSIRO2, RCP8.5
                    np.array(
                        [[1] * 3 + [0] * 1, [1] * 3 + [0] * 1]
                    ),  # EC-EARTH RCMs, RCP8.5
                    np.array(
                        [[0.5] * 2 + [0] * 2, [0.5] * 2 + [0] * 2]
                    ),  # GFDL-ESM2G family, RCP4.5
                    np.array([[1] * 4, [1] * 4]),  # GFDL-ESM2M family, RCP4.5
                )
            ),
            "GCM-True": np.concatenate(
                (
                    np.array(
                        [
                            [0.02777, 0.04166, 0.05555, 0.08333],
                            [0.02777, 0.04166, 0.05555, 0.08333],
                            [0.02777, 0.04166, 0.05555, 0.08333],
                            [0.08333, 0.125, 0.16666, 0.25],
                            [0.08333, 0, 0, 0],
                        ]
                    ),  # CanESM2 family, RCP4.5
                    np.array(
                        [
                            [0.00854, 0.00925, 0.00925, 0.02777],
                            [0.00854, 0.00925, 0.00925, 0.02777],
                            [0.00854, 0.00925, 0.00925, 0.02777],
                            [0.02564, 0.02777, 0.02777, 0.08333],
                            [0.02564, 0, 0, 0],
                        ]
                    ),  # CanESM2 family, RCP8.5
                    np.repeat(
                        np.array([[0.02564, 0.02777, 0.02777, 0.08333]]), 10, 0
                    ),  # ClimEx
                    np.repeat(
                        np.array([[0.05, 0.05, 0.06666, 0]]), 5, 0
                    ),  # CSIRO-Mk6, RCP4.5
                    np.array([[0.33333, 0.33333, 0.33333, 0]]),  # CSIRO2, RCP8.5
                    np.array(
                        [[0.16666, 0.16666, 0.16666, 0], [0.16666, 0.16666, 0.16666, 0]]
                    ),  # EC-EARTH RCMs, RCP8.5
                    np.array(
                        [[0.125, 0.125, 0, 0], [0.125, 0.125, 0, 0]]
                    ),  # GFDL-ESM2G family, RCP4.5
                    np.array(
                        [[0.125, 0.125, 0.16666, 0.25], [0.125, 0.125, 0.16666, 0.25]]
                    ),  # GFDL-ESM2M family, RCP4.5
                )
            ),
            "GCM-False": np.concatenate(
                (
                    np.array(
                        [
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.0625] * 1 + [0.07142] * 3,
                            [0.0625] * 1 + [0] * 3,
                        ]
                    ),  # CanESM2 family, RCP4.5
                    np.array(
                        [
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.0625] * 1 + [0.07142] * 3,
                            [0.0625] * 1 + [0] * 3,
                        ]
                    ),  # CanESM2 family, RCP8.5
                    np.repeat(
                        np.array([[0.0625] * 1 + [0.07142] * 3]), 10, 0
                    ),  # ClimEx
                    np.repeat(
                        np.array([[0.2] * 3 + [0] * 1]), 5, 0
                    ),  # CSIRO-Mk6, RCP4.5
                    np.array([[1] * 3 + [0] * 1]),  # CSIRO2, RCP8.5
                    np.array(
                        [[0.5] * 3 + [0] * 1, [0.5] * 3 + [0] * 1]
                    ),  # EC-EARTH RCMs, RCP8.5
                    np.array(
                        [[0.5] * 2 + [0] * 2, [0.5] * 2 + [0] * 2]
                    ),  # GFDL-ESM2G family, RCP4.5
                    np.array([[0.5] * 4, [0.5] * 4]),  # GFDL-ESM2M family, RCP4.5
                )
            ),
            "institution-True": np.concatenate(
                (
                    np.array(
                        [
                            [0.03703, 0.05555, 0.05555, 0.08333],
                            [0.03703, 0.05555, 0.05555, 0.08333],
                            [0.03703, 0.05555, 0.05555, 0.08333],
                            [0.11111, 0.16666, 0.16666, 0.25],
                            [0.11111, 0, 0, 0],
                        ]
                    ),  # CanESM2 family, RCP4.5
                    np.array(
                        [
                            [0.00854, 0.00925, 0.00925, 0.02777],
                            [0.00854, 0.00925, 0.00925, 0.02777],
                            [0.00854, 0.00925, 0.00925, 0.02777],
                            [0.02564, 0.02777, 0.02777, 0.08333],
                            [0.02564, 0, 0, 0],
                        ]
                    ),  # CanESM2 family, RCP8.5
                    np.repeat(
                        np.array([[0.02564, 0.02777, 0.02777, 0.08333]]), 10, 0
                    ),  # ClimEx
                    np.repeat(
                        np.array([[0.06666, 0.06666, 0.06666, 0]]), 5, 0
                    ),  # CSIRO-Mk6, RCP4.5
                    np.array([[0.33333, 0.33333, 0.33333, 0]]),  # CSIRO2, RCP8.5
                    np.array(
                        [[0.16666, 0.16666, 0.16666, 0], [0.16666, 0.16666, 0.16666, 0]]
                    ),  # EC-EARTH RCMs, RCP8.5
                    np.array(
                        [[0.08333, 0.08333, 0, 0], [0.08333, 0.08333, 0, 0]]
                    ),  # GFDL-ESM2G family, RCP4.5
                    np.array(
                        [
                            [0.08333, 0.08333, 0.16666, 0.25],
                            [0.08333, 0.08333, 0.16666, 0.25],
                        ]
                    ),  # GFDL-ESM2M family, RCP4.5
                )
            ),
            "institution-False": np.concatenate(
                (
                    np.array(
                        [
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.0625] * 1 + [0.07142] * 3,
                            [0.0625] * 1 + [0] * 3,
                        ]
                    ),  # CanESM2 family, RCP4.5
                    np.array(
                        [
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.02083] * 1 + [0.02380] * 3,
                            [0.0625] * 1 + [0.07142] * 3,
                            [0.0625] * 1 + [0] * 3,
                        ]
                    ),  # CanESM2 family, RCP8.5
                    np.repeat(
                        np.array([[0.0625] * 1 + [0.07142] * 3]), 10, 0
                    ),  # ClimEx
                    np.repeat(
                        np.array([[0.1] * 3 + [0] * 1]), 5, 0
                    ),  # CSIRO-Mk6, RCP4.5
                    np.array([[0.5] * 3 + [0] * 1]),  # CSIRO2, RCP8.5
                    np.array(
                        [[0.5] * 3 + [0] * 1, [0.5] * 3 + [0] * 1]
                    ),  # EC-EARTH RCMs, RCP8.5
                    np.array(
                        [[0.25] * 2 + [0] * 2, [0.25] * 2 + [0] * 2]
                    ),  # GFDL-ESM2G family, RCP4.5
                    np.array(
                        [[0.25] * 2 + [0.5] * 2, [0.25] * 2 + [0.5] * 2]
                    ),  # GFDL-ESM2M family, RCP4.5
                )
            ),
        }

        answer = answers[f"{independence_level}-{exp_weights}"]
        if skipna:
            answer = answer[:, 0]

        return answer

    @pytest.mark.parametrize(
        "independence_level, exp_weights, skipna",
        [
            ("model", True, True),
            ("model", False, True),
            ("GCM", True, True),
            ("GCM", False, True),
            ("institution", True, True),
            ("institution", False, True),
            ("model", True, False),
            ("model", False, False),
            ("GCM", True, False),
            ("GCM", False, False),
            ("institution", True, False),
            ("institution", False, False),
        ],
    )
    def test_generate_weights(self, independence_level, exp_weights, skipna):
        out = xs.generate_weights(
            self.ens,
            independence_level=independence_level,
            balance_experiments=exp_weights,
            skipna=skipna,
        )

        answer = self.make_answer(independence_level, exp_weights, skipna)
        np.testing.assert_array_almost_equal(out, answer, decimal=4)

    def test_changing_horizon(self):
        ens = deepcopy(self.ens)
        ds = ens["CCCma-CanESM2-rcp45-r1i1p1-CanESM2"]
        hz = xr.DataArray(
            ["1981-2010", "2041-2070", "+2C", "not_the_same"], dims="horizon"
        )
        ens["CCCma-CanESM2-rcp45-r1i1p1-CanESM2"] = ds.assign_coords({"horizon": hz})

        out = xs.generate_weights(ens, skipna=False)
        assert len(out.horizon) == 5
        # Should all be 0s, except for CCCma-CanESM2-rcp45-r1i1p1-CanESM2
        np.testing.assert_array_equal(
            out.sel(
                realization="CCCma-CanESM2-rcp45-r1i1p1-CanESM2", horizon="not_the_same"
            ),
            1,
        )
        assert all(
            out.sel(
                realization=~out.realization.isin(
                    ["CCCma-CanESM2-rcp45-r1i1p1-CanESM2"]
                ),
                horizon="not_the_same",
            )
            == 0
        )

    @pytest.mark.parametrize(
        "standardize, skipna, attribute_weights",
        [
            (True, True, None),
            (True, False, None),
            (False, True, None),
            (
                True,
                True,
                {"institution": {"CCCma": 2, "ECMWF": 0, "GFDL": 5}},
            ),
        ],
    )
    def test_standardize(self, standardize, skipna, attribute_weights):
        if attribute_weights:
            datasets = self.ens_rcm
        else:
            datasets = self.ens
        out = xs.generate_weights(
            datasets,
            standardize=standardize,
            skipna=skipna,
            attribute_weights=attribute_weights,
        )
        if standardize:
            np.testing.assert_allclose(out.sum(), 1 if skipna else 4)

        else:
            np.testing.assert_allclose(out.sum(), 10)

    def test_spatial(self, datablock_3d):
        a = datablock_3d(
            np.random.rand(10, 10, 10), "tas", "lon", 0, "lat", 0, as_dataset=True
        )
        a.attrs = {
            "cat:institution": "CCCma",
            "cat:source": "CanESM5",
            "cat:experiment": "ssp585",
            "cat:member": "r1i1p1f1",
        }
        b = datablock_3d(
            np.random.rand(10, 10, 10), "tas", "lon", 0, "lat", 0, as_dataset=True
        )
        b.attrs = {
            "cat:institution": "CCCma",
            "cat:source": "CanESM5",
            "cat:experiment": "ssp585",
            "cat:member": "r2i1p1f1",
        }
        c = datablock_3d(
            np.random.rand(10, 10, 10), "tas", "lon", 0, "lat", 0, as_dataset=True
        )
        c.attrs = {
            "cat:institution": "CSIRO",
            "cat:source": "ACCESS-ESM1-5",
            "cat:experiment": "ssp585",
            "cat:member": "r1i1p1f1",
        }
        ens = [a, b, c]

        np.testing.assert_allclose(
            xs.generate_weights(ens, skipna=True), [0.5, 0.5, 1.0]
        )
        with pytest.warns(UserWarning, match="Dataset 0 has dimensions that are not "):
            np.testing.assert_allclose(
                xs.generate_weights(ens, skipna=False),
                [[0.5] * 10, [0.5] * 10, [1.0] * 10],
            )

    def test_errors(self):
        # Bad input
        with pytest.raises(ValueError, match="'independence_level' should be between"):
            xs.generate_weights(self.ens, independence_level="foo")

        ens2 = deepcopy(self.ens)
        ens2["CCCma-CanESM2-rcp45-r1i1p1-CanESM2"] = ens2[
            "CCCma-CanESM2-rcp45-r1i1p1-CanESM2"
        ].expand_dims("time")
        with pytest.raises(
            ValueError,
            match="Expected either 'time' or 'horizon' as an extra dimension",
        ):
            xs.generate_weights(ens2, skipna=False)
        xs.generate_weights(ens2, skipna=True)  # Should work

        # Required attributes
        ens2 = deepcopy(self.ens)
        ens2["CCCma-CanESM2-rcp45-r1i1p1-CanESM2"].attrs["cat:source"] = None
        with pytest.raises(
            ValueError,
            match="The 'cat:source' or 'cat:driving_model' attribute is missing",
        ):
            xs.generate_weights(ens2)
        ens2 = deepcopy(self.ens)
        ens2["CCCma-CanESM2-rcp45-r1i1p1-CanESM2"].attrs["cat:experiment"] = None
        with pytest.raises(
            ValueError, match="The 'cat:experiment' attribute is missing"
        ):
            xs.generate_weights(ens2, balance_experiments=True)
        ens2 = deepcopy(self.ens)
        ens2["CCCma-CanESM2-rcp45-r1i1p1-CanESM2"].attrs["cat:institution"] = None
        with pytest.raises(
            ValueError, match="The 'cat:institution' attribute is missing"
        ):
            xs.generate_weights(ens2, independence_level="institution")

        # Optional, but recommended attributes
        ens2 = deepcopy(self.ens)
        for e in ens2:
            ens2[e].attrs["cat:experiment"] = None
        with pytest.warns(
            UserWarning,
            match="The 'cat:experiment' attribute is missing from all datasets",
        ):
            xs.generate_weights(ens2, balance_experiments=False)
        ens2 = deepcopy(self.ens)
        ens2["CCCma-CanESM2-rcp45-r1i1p1-CanESM2"].attrs["cat:member"] = None
        with pytest.warns(
            UserWarning,
            match="The 'cat:member' attribute is inconsistent across datasets.",
        ):
            xs.generate_weights(ens2)

    @staticmethod
    def answer_attribute_weight(can45, can85, clim, csi45, csi85, ec85, gfd45):
        ans = np.concatenate(
            (
                np.array(
                    np.split(
                        np.array(
                            [
                                0.08333,
                                0.25,
                                0.25,
                                0.25,
                                0.25,
                            ]
                        ),
                        5,
                    )
                )
                * can45,  # CanESM2 family, RCP4.5
                np.array(
                    np.split(
                        np.array(
                            [
                                0.08333,
                                0.25,
                                0.25,
                                0.25,
                                0.25,
                            ]
                        ),
                        5,
                    )
                )
                * can85,  # CanESM2 family, RCP8.5
                np.array(np.split(np.repeat(np.array(0.08333), 10, 0), 10))
                * clim,  # ClimEx
                np.array(np.split(np.repeat(0.2, 5, 0), 5))
                * csi45,  # CSIRO-Mk6, RCP4.5
                np.array([1], ndmin=2) * csi85,  # CSIRO2, RCP8.5
                np.array(np.split(np.array([1, 1]), 2)) * ec85,  # EC-EARTH RCMs, RCP8.5
                np.array(np.split(np.array([0.5, 0.5]), 2))
                * gfd45,  # GFDL-ESM2G family, RCP4.5
                np.array(np.split(np.array([1, 1]), 2))
                * gfd45,  # GFDL-ESM2M family, RCP4.5
            )
        )
        return np.squeeze(ans)

    @pytest.mark.parametrize(
        "coefs, weights",
        [
            (
                {
                    "can45": 2,
                    "can85": 2,
                    "clim": 2,
                    "csi45": 3,
                    "csi85": 3,
                    "ec85": 0,
                    "gfd45": 5,
                },
                {"institution": {"CCCma": 2, "CSIRO-QCCCE": 3, "ECMWF": 0, "GFDL": 5}},
            ),
            (
                {
                    "can45": 2,
                    "can85": 2,
                    "clim": 2,
                    "csi45": 1,
                    "csi85": 1,
                    "ec85": 1,
                    "gfd45": 1,
                },
                {"institution": {"CCCma": 2, "others": 1}},
            ),
            (
                {
                    "can45": 2,
                    "can85": 1,
                    "clim": 1,
                    "csi45": 2,
                    "csi85": 1,
                    "ec85": 1,
                    "gfd45": 2,
                },
                {"experiment": {"rcp45": 2, "rcp85": 1}},
            ),
            (
                {
                    "can45": [0, 0, 1, 1],
                    "can85": [0, 1, 0, 1],
                    "clim": [0, 1, 0, 1],
                    "csi45": [0, 0, 1, 1],
                    "csi85": [0, 1, 0, 1],
                    "ec85": [0, 1, 0, 1],
                    "gfd45": [0, 0, 1, 1],
                },
                {
                    "experiment": xr.DataArray(
                        data=np.array([[0, 0], [0, 1], [1, 0], [1, 1]]),
                        dims=["horizon", "experiment"],
                        coords=dict(
                            horizon=["1981-2010", "2041-2070", "+2C", "+4C"],
                            experiment=["rcp45", "rcp85"],
                        ),
                        attrs=dict(
                            description="Experiment weight through horizons",
                            units="",
                        ),
                    )
                },
            ),
            (
                {
                    "can45": [0, 0, 1, 1],
                    "can85": [0, 1, 0, 1],
                    "clim": [0, 1, 0, 1],
                    "csi45": [0, 0, 1, 1],
                    "csi85": [0, 1, 0, 1],
                    "ec85": [0, 1, 0, 1],
                    "gfd45": [0, 0, 1, 1],
                },
                {
                    "experiment": xr.DataArray(
                        data=np.array([[0, 0], [0, 1], [1, 0], [1, 1]]),
                        dims=["horizon", "experiment"],
                        coords=dict(
                            horizon=["1981-2010", "2041-2070", "+2C", "+4C"],
                            experiment=["rcp45", "others"],
                        ),
                        attrs=dict(
                            description="Experiment weight through horizons",
                            units="",
                        ),
                    )
                },
            ),
        ],
    )
    def test_attribute_weight(self, coefs, weights):
        # only test for RCMs

        # generate ans
        ans = self.answer_attribute_weight(**coefs)
        ind = [
            index
            for index, value in enumerate(list(self.ens.keys()))
            if value in list(self.ens_rcm.keys())
        ]
        ans_rcms = ans[ind]
        if len(ans_rcms.shape) > 1:
            ans_rcms = ans_rcms.transpose()
        out = xs.generate_weights(self.ens_rcm, attribute_weights=weights)
        np.testing.assert_array_almost_equal(out, ans_rcms, decimal=4)

    def test_attribute_weight_error(self):
        # Required attributes
        with pytest.raises(
            ValueError,
            match="Attribute_weights should be dict or xr.DataArray.",
        ):
            xs.generate_weights(
                self.ens_rcm, attribute_weights={"experiment": [1, 2, 3]}
            )
        with pytest.raises(
            ValueError,
            match="The test attribute is missing from some simulations.",
        ):
            xs.generate_weights(
                self.ens_rcm, attribute_weights={"test": {"CCCma": 2, "others": 1}}
            )
        with pytest.raises(
            ValueError,
            match="The institution ECMWF or others are not in the attribute_weights dict.",
        ):
            xs.generate_weights(
                self.ens_rcm, attribute_weights={"institution": {"CCCma": 2, "GFDL": 1}}
            )
        with pytest.raises(
            ValueError, match="experiment is not in the xr.DataArray coords."
        ):
            xs.generate_weights(
                self.ens_rcm,
                attribute_weights={
                    "experiment": xr.DataArray(
                        data=np.array([[0, 0], [0, 1], [1, 0], [1, 1]]),
                        dims=["horizon", "test"],
                        coords=dict(
                            horizon=["1981-2010", "2041-2070", "+2C", "+4C"],
                            test=["rcp45", "rcp85"],
                        ),
                    )
                },
            )
        with pytest.raises(
            ValueError,
            match="The experiment DataArray has more than one coord dimension to apply weights",
        ):
            xs.generate_weights(
                self.ens_rcm,
                attribute_weights={
                    "experiment": xr.DataArray(
                        data=np.array(
                            [
                                [[0, 0], [0, 1]],
                                [[0, 1], [2, 1]],
                                [[1, 0], [1, 2]],
                                [[1, 1], [2, 2]],
                            ]
                        ),
                        dims=["horizon", "experiment", "test"],
                        coords=dict(
                            horizon=["1981-2010", "2041-2070", "+2C", "+4C"],
                            experiment=["rcp45", "rcp85"],
                            test=[0, 1],
                        ),
                    )
                },
            )
        with pytest.raises(
            ValueError,
            match="The experiment rcp85 or others are not in the attribute_weights datarray coords.",
        ):
            xs.generate_weights(
                self.ens_rcm,
                attribute_weights={
                    "experiment": xr.DataArray(
                        data=np.array([[0], [0], [1], [1]]),
                        dims=["horizon", "experiment"],
                        coords=dict(
                            horizon=["1981-2010", "2041-2070", "+2C", "+4C"],
                            experiment=["rcp45"],
                        ),
                    )
                },
            )
            # mismatch RCMs & GCMs
            with pytest.raises(
                NotImplementedError,
                match="Management of RCM and GCM in same datasets dictionary not "
                "yet implemented with attribute_weights.",
            ):
                xs.generate_weights(
                    self.ens,
                    attribute_weights={"institution": {"CCCma": 2, "others": 1}},
                )
            # warning mismatch attribute_weights & independance_level
            with pytest.warns(
                UserWarning,
                match="The institution weights do not match the model independance_level",
            ):
                xs.generate_weights(
                    self.ens_rcm,
                    attribute_weights={"institution": {"CCCma": 2, "others": 1}},
                )
            with pytest.warns(
                UserWarning,
                match="Key experiment given in attribute_weights without argument balance_experiments=True",
            ):
                xs.generate_weights(
                    self.ens_rcm,
                    attribute_weights={"experiment": {"rcp45": 2, "rcp85": 1}},
                )


class TestEnsemblePartition:
    @pytest.mark.skipif(xe is None, reason="xesmf needed for testing regrdding")
    def test_build_partition_data(self, samplecat, tmp_path):
        # test subset
        datasets = samplecat.search(variable="tas", member="r1i1p1f1").to_dataset_dict(
            xarray_open_kwargs={"engine": "h5netcdf"}
        )
        ds = xs.ensembles.build_partition_data(
            datasets=datasets,
            partition_dim=["source", "experiment"],
            subset_kw=dict(name="mtl", method="gridpoint", lat=[45.0], lon=[-74]),
            rename_dict={"source": "new-name"},
        )

        assert ds.sizes == {"time": 730, "scenario": 4, "new-name": 1}
        assert ds.attrs["cat:processing_level"] == "partition-ensemble"

        # test regrid
        ds_grid = xe.util.cf_grid_2d(-75, -74, 0.25, 45, 48, 0.55)
        datasets = samplecat.search(variable="tas", member="r1i1p1f1").to_dataset_dict(
            xarray_open_kwargs={"engine": "h5netcdf"}
        )
        ds = xs.ensembles.build_partition_data(
            datasets=datasets,
            regrid_kw=dict(ds_grid=ds_grid, weights_location=tmp_path),
            to_level="test",
        )

        assert ds.sizes == {
            "scenario": 4,
            "model": 1,
            "time": 730,
            "lat": 5,
            "lon": 4,
        }
        assert [i for i in ds.data_vars] == ["tas"]
        assert ds.attrs["cat:processing_level"] == "test"

    def test_partition_from_catalog(self, samplecat):
        datasets = samplecat.search(variable="tas", member="r1i1p1f1")
        ds_from_dict = xs.ensembles.build_partition_data(
            datasets=datasets.to_dataset_dict(
                xarray_open_kwargs={"engine": "h5netcdf"}
            ),
            partition_dim=["source", "experiment"],
            subset_kw=dict(name="mtl", method="gridpoint", lat=[45.0], lon=[-74]),
        )

        ds_from_cat = xs.ensembles.build_partition_data(
            datasets=datasets,
            partition_dim=["source", "experiment"],
            subset_kw=dict(name="mtl", method="gridpoint", lat=[45.0], lon=[-74]),
            to_dataset_kw=dict(xarray_open_kwargs={"engine": "h5netcdf"}),
        )
        # fix order
        ds_from_cat = ds_from_cat[["time", "model", "scenario", "tas"]]
        ds_from_cat["tas"] = ds_from_cat["tas"].transpose("scenario", "model", "time")

        assert ds_from_dict.equals(ds_from_cat)

    def test_realization_partition(self, samplecat):

        datasets = samplecat.search(variable="tas").to_dataset_dict(
            xarray_open_kwargs={"engine": "h5netcdf"}
        )
        ds = xs.ensembles.build_partition_data(
            datasets=datasets,
            partition_dim=["realization", "experiment"],
            subset_kw=dict(name="mtl", method="gridpoint", lat=[45.0], lon=[-74]),
        )

        assert "NCC_NorESM2-MM_r1i1p1f1" in ds.model.values
        assert ds.sizes == {"time": 730, "scenario": 4, "model": 2}


class TestReduceEnsemble:
    def test_with_criteria(self):
        ds = xr.open_dataset(nimbus().fetch("EnsembleReduce/TestEnsReduceCriteria.nc"))
        selected, clusters, fig_data = xs.reduce_ensemble(
            ds["data"], method="kmeans", max_clusters=3
        )
        assert selected.shape == (3,)
        np.testing.assert_array_equal(selected, [4, 7, 23])
        assert len(clusters) == 3
        assert fig_data is not None

    @pytest.mark.parametrize("horizon", ["1981-2010", "2021-2050"])
    def test_without_criteria(self, horizon):
        datasets = {
            "ACCESS": "EnsembleStats/BCCAQv2+ANUSPLIN300_ACCESS1-0_historical+rcp45_r1i1p1_1950-2100_tg_mean_YS.nc",
            "BNU-ESM": "EnsembleStats/BCCAQv2+ANUSPLIN300_BNU-ESM_historical+rcp45_r1i1p1_1950-2100_tg_mean_YS.nc",
            "CCSM4-r1": "EnsembleStats/BCCAQv2+ANUSPLIN300_CCSM4_historical+rcp45_r1i1p1_1950-2100_tg_mean_YS.nc",
            "CCSM4-r2": "EnsembleStats/BCCAQv2+ANUSPLIN300_CCSM4_historical+rcp45_r2i1p1_1950-2100_tg_mean_YS.nc",
            "CNRM-CM5": "EnsembleStats/BCCAQv2+ANUSPLIN300_CNRM-CM5_historical+rcp45_r1i1p1_1970-2050_tg_mean_YS.nc",
        }
        for d in datasets:
            ds = xr.open_dataset(nimbus().fetch(datasets[d])).isel(
                lon=slice(0, 4), lat=slice(0, 4)
            )
            ds = xs.climatological_op(
                ds,
                op="mean",
                window=30,
                periods=[["1981", "2010"], ["2021", "2050"]],
                horizons_as_dim=True,
            ).drop_vars("time")
            datasets[d] = xs.compute_deltas(ds, reference_horizon="1981-2010")

        selected, clusters, fig_data = xs.reduce_ensemble(
            datasets, method="kkz", horizons=[horizon], num_select=4
        )
        assert selected.shape == (4,)
        answer = (
            ["ACCESS", "BNU-ESM", "CNRM-CM5", "CCSM4-r1"]
            if horizon == "2021-2050"
            else ["ACCESS", "BNU-ESM", "CCSM4-r1", "CCSM4-r2"]
        )
        np.testing.assert_array_equal(selected, answer)
        assert clusters == {}
        assert fig_data == {}

    def test_errors(self):
        ds = xr.open_dataset(nimbus().fetch("EnsembleReduce/TestEnsReduceCriteria.nc"))
        with pytest.raises(
            ValueError, match="Data must have a 'horizon' dimension to be subsetted."
        ):
            xs.reduce_ensemble(
                ds["data"], method="kmeans", horizons=["1981-2010"], max_clusters=3
            )



from pathlib import Path

import pandas as pd
import xarray as xr
from conftest import SAMPLES_DIR

from xscen import catalog, extract


def test_subset_file_coverage():
    df = pd.DataFrame.from_records(
        [
            {"id": "A", "date_start": "1950-01-01", "date_end": "1959-12-31"},
            {"id": "A", "date_start": "1960-01-01", "date_end": "1969-12-31"},
        ]
    )
    df["date_start"] = pd.to_datetime(df.date_start)
    df["date_end"] = pd.to_datetime(df.date_end)

    # Ok
    pd.testing.assert_frame_equal(
        catalog.subset_file_coverage(df, [1951, 1970], coverage=0.8), df
    )

    # Insufficient coverage (no files)
    pd.testing.assert_frame_equal(
        catalog.subset_file_coverage(df, [1975, 1986], coverage=0.8),
        pd.DataFrame(columns=df.columns),
    )

    # Insufficient coverage (partial files)
    pd.testing.assert_frame_equal(
        catalog.subset_file_coverage(df, [1951, 1976], coverage=0.8),
        pd.DataFrame(columns=df.columns),
    )


def test_xrfreq_fix():
    cat = catalog.DataCatalog(SAMPLES_DIR.parent / "pangeo-cmip6.json")
    assert set(cat.df.xrfreq) == {"3h", "D", "fx"}


class TestCopyFiles:
    def test_flat(self, samplecat, tmp_path):
        newcat = samplecat.copy_files(tmp_path, flat=True)
        assert len(list(tmp_path.glob("*.nc"))) == len(newcat.df)

    def test_inplace(self, samplecat, tmp_path):
        dsid, scat = extract.search_data_catalogs(
            data_catalogs=[samplecat],
            variables_and_freqs={"tas": "MS"},
            allow_resampling=True,
            other_search_criteria={
                "experiment": "ssp585",
                "source": "NorESM.*",
                "member": "r1i1p1f1",
            },
        ).popitem()
        scat.copy_files(tmp_path, inplace=True)
        assert len(list(tmp_path.glob("*.nc"))) == len(scat.df)

        _, ds = extract.extract_dataset(scat).popitem()
        frq = xr.infer_freq(ds.time)
        assert frq == "MS"

    def test_zipunzip(self, samplecat, tmp_path):
        dsid, scat = extract.search_data_catalogs(
            data_catalogs=[samplecat],
            variables_and_freqs={"tas": "D"},
            allow_resampling=True,
            other_search_criteria={
                "experiment": "ssp585",
                "source": "NorESM.*",
                "member": "r1i1p1f1",
            },
        ).popitem()
        _, ds = extract.extract_dataset(scat).popitem()
        ds.to_zarr(tmp_path / "temp.zarr")
        scat.esmcat.df.loc[0, "path"] = tmp_path / "temp.zarr"

        rz = tmp_path / "zipped"
        rz.mkdir()
        scat_z = scat.copy_files(rz, zipzarr=True)
        f = Path(scat_z.df.path.iloc[0])
        assert f.suffix == ".zip"
        assert f.parent.name == rz.name
        assert f.is_file()

        ru = tmp_path / "unzipped"
        ru.mkdir()
        scat_uz = scat.copy_files(ru, unzip=True)
        f = Path(scat_uz.df.path.iloc[0])
        assert f.suffix == ".zarr"
        assert f.parent.name == ru.name
        assert f.is_dir()



"""A collection of various convenience objects and functions to use in scripts."""

import logging
import mimetypes
import os
import shutil as sh
import signal
import smtplib
import sys
import time
import warnings
from contextlib import contextmanager
from email.message import EmailMessage
from io import BytesIO
from pathlib import Path
from traceback import format_exception

import xarray as xr
from matplotlib.figure import Figure

from .catalog import ProjectCatalog
from .config import parse_config
from .utils import get_cat_attrs

logger = logging.getLogger(__name__)

__all__ = [
    "TimeoutException",
    "measure_time",
    "move_and_delete",
    "save_and_update",
    "send_mail",
    "send_mail_on_exit",
    "skippable",
    "timeout",
]


@parse_config
def send_mail(
    *,
    subject: str,
    msg: str,
    to: str | None = None,
    server: str = "127.0.0.1",
    port: int = 25,
    attachments: None | (
        list[tuple[str, Figure | os.PathLike] | Figure | os.PathLike]
    ) = None,
) -> None:
    """Send email.

    Email a single address through a login-less SMTP server.
    The default values of server and port should work out-of-the-box on Ouranos's systems.

    Parameters
    ----------
    subject: str
      Subject line.
    msg: str
      Main content of the email. Can be UTF-8 and multi-line.
    to: str, optional
      Email address to which send the email. If None (default), the email is sent to "{os.getlogin()}@{os.uname().nodename}".
      On unix systems simply put your real email address in `$HOME/.forward` to receive the emails sent to this local address.
    server : str
      SMTP server url. Defaults to 127.0.0.1, the local host. This function does not try to log-in.
    port: int
      Port of the SMTP service on the server. Defaults to 25, which is usually the default port on unix-like systems.
    attachments : list of paths or matplotlib figures or tuples of a string and a path or figure, optional
      List of files to attach to the email.
      Elements of the list can be paths, the mimetypes of those is guessed and the files are read and sent.
      Elements can also be matplotlib Figures which are send as png image (savefig) with names like "Figure00.png".
      Finally, elements can be tuples of a filename to use in the email and the attachment, handled as above.

    Returns
    -------
    None
    """
    # Inspired by https://betterprogramming.pub/how-to-send-emails-with-attachments-using-python-dd37c4b6a7fd
    email = EmailMessage()
    email["Subject"] = subject
    email["From"] = f"{os.getlogin()}@{os.uname().nodename}"
    email["To"] = to or f"{os.getlogin()}@{os.uname().nodename}"
    email.set_content(msg)

    for i, att in enumerate(attachments or []):
        fname = None
        if isinstance(att, tuple):
            fname, att = att
        if isinstance(att, Figure):
            data = BytesIO()
            att.savefig(data, format="png")
            data.seek(0)
            email.add_attachment(
                data.read(),
                maintype="image",
                subtype="png",
                filename=fname or f"Figure{i:02d}.png",
            )
        else:  # a path
            attpath = Path(att)
            ctype, encoding = mimetypes.guess_type(attpath)
            if ctype is None or encoding is not None:
                ctype = "application/octet-stream"
            maintype, subtype = ctype.split("/", 1)

            with attpath.open("rb") as fp:
                email.add_attachment(
                    fp.read(),
                    maintype=maintype,
                    subtype=subtype,
                    filename=fname or attpath.name,
                )

    with smtplib.SMTP(host=server, port=port) as SMTP:
        SMTP.send_message(email)


class ExitWatcher:
    """An object that watches system exits and exceptions before the python exits."""

    def __init__(self):
        self.code = None
        self.error = None
        self.hooked = False

    def hook(self):
        """Hooks the watcher to the system by monkeypatching `sys` with its own methods."""
        if not self.hooked:
            self.orig_exit = sys.exit
            self.orig_excepthook = sys.excepthook
            sys.exit = self.exit
            sys.excepthook = self.err_handler
            self.hooked = True
        else:
            warnings.warn("Exit hooks have already been overridden.")

    def unhook(self):
        if self.hooked:
            sys.exit = self.orig_exit
            sys.excepthook = self.orig_excepthook
        else:
            raise ValueError("Exit hooks were not overridden. Cannot unhook.")

    def exit(self, code=0):
        self.code = code
        self.orig_exit(code)

    def err_handler(self, *exc_info):
        self.error = exc_info
        self.orig_excepthook(*exc_info)


exit_watcher = ExitWatcher()
exit_watcher.hook()


@parse_config
def send_mail_on_exit(
    *,
    subject: str | None = None,
    msg_ok: str | None = None,
    msg_err: str | None = None,
    on_error_only: bool = False,
    skip_ctrlc: bool = True,
    **mail_kwargs,
) -> None:
    """Send an email with content depending on how the system exited.

    This function is best used by registering it with `atexit`. Calls :py:func:`send_mail`.

    Parameters
    ----------
    subject : str, optional
        Email subject. Will be appended by "Success", "No errors" or "Failure" depending
        on how the system exits.
    msg_ok : str, optional
        Content of the email if the system exists successfully.
    msg_err : str, optional
        Content of the email id the system exists with a non-zero code or with an error.
        The message will be appended by the exit code or with the error traceback.
    on_error_only : boolean
        Whether to only send an email on a non-zero/error exit.
    skip_ctrlc : boolean
        If True (default), exiting with a KeyboardInterrupt will not send an email.
    mail_kwargs
        Other arguments passed to :py:func:`send_mail`.
        The `to` argument is necessary for this function to work.

    Example
    -------
    Send an eamil titled "Woups" upon non-successful program exit. We assume the `to`
    field was given in the config.

    >>> import atexit
    >>> atexit.register(send_mail_on_exit, subject="Woups", on_error_only=True)
    """
    subject = subject or "Workflow"
    msg_err = msg_err or "Workflow exited with some errors."
    if (
        not on_error_only
        and exit_watcher.error is None
        and exit_watcher.code in [None, 0]
    ):
        send_mail(
            subject=subject + " - Success",
            msg=msg_ok or "Workflow exited successfully.",
            **mail_kwargs,
        )
    elif exit_watcher.error is None and (exit_watcher.code or 0) > 0:
        send_mail(
            subject=subject + " - No errors",
            msg=f"{msg_err}\nSystem exited with code {exit_watcher.code}.",
            **mail_kwargs,
        )
    elif exit_watcher.error is not None and (
        exit_watcher.error[0] is not KeyboardInterrupt or not skip_ctrlc
    ):
        tb = "".join(format_exception(*exit_watcher.error))
        msg_err = f"{msg_err}\n\n{tb}"
        send_mail(subject=subject + " - Failure", msg=msg_err, **mail_kwargs)


@parse_config
class measure_time:
    """Context for timing a code block.

    Parameters
    ----------
    name : str, optional
      A name to give to the block being timed, for meaningful logging.
    cpu : boolean
      If True, the CPU time is also measured and logged.
    logger : logging.Logger, optional
      The logger object to use when sending Info messages with the measured time.
      Defaults to a logger from this module.
    """

    def __init__(
        self,
        name: str | None = None,
        cpu: bool = False,
        logger: logging.Logger = logger,
    ):
        self.name = name or ""
        self.cpu = cpu
        self.logger = logger

    def __enter__(self):  # noqa: D105
        self.start = time.perf_counter()
        self.start_cpu = time.process_time()
        msg = f"Started process {self.name}."
        self.logger.info(msg)
        return

    def __exit__(self, *args, **kwargs):  # noqa: D105
        elapsed = time.perf_counter() - self.start
        elapsed_cpu = time.process_time() - self.start_cpu
        occ = elapsed_cpu / elapsed
        s = f"Process {self.name} done in {elapsed:.02f} s"
        if self.cpu:
            s += f" and used {elapsed_cpu:.02f} of cpu time ({occ:.1%} % occupancy)."

        self.logger.info(s)


# FIXME: This should be written as "TimeoutError"
class TimeoutException(Exception):  # noqa: N818
    """An exception raised with a timeout occurs."""

    def __init__(self, seconds: int, task: str = "", **kwargs):
        self.msg = f"Task {task} timed out after {seconds} seconds"
        super().__init__(self.msg, **kwargs)


@contextmanager
def timeout(seconds: int, task: str = ""):
    """Timeout context manager.

    Only one can be used at a time, this is not multithread-safe : it cannot be used in
    another thread than the main one, but multithreading can be used in parallel.

    Parameters
    ----------
    seconds : int
      Number of seconds after which the context exits with a TimeoutException.
      If None or negative, no timeout is set and this context does nothing.
    task : str, optional
      A name to give to the task, allowing a more meaningful exception.
    """
    if seconds is None or seconds <= 0:
        yield
    else:

        # FIXME: These variables are not used
        def _timeout_handler(signum, frame):  # noqa: F841
            raise TimeoutException(seconds, task)

        old_handler = signal.signal(signal.SIGALRM, _timeout_handler)
        signal.alarm(seconds)
        try:
            yield
        finally:
            signal.alarm(0)
            signal.signal(signal.SIGALRM, old_handler)


@contextmanager
def skippable(seconds: int = 2, task: str = "", logger: logging.Logger | None = None):
    """Skippable context manager.

    When CTRL-C (SIGINT, KeyboardInterrupt) is sent within the context,
    this catches it, prints to the log and gives a timeout during which a subsequent
    interruption will stop the script. Otherwise, the context exits normally.

    This is meant to be used within a loop so that we can skip some iterations:

    .. code-block:: python

        for i in iterable:
            with skippable(2, i):
                some_skippable_code()

    Parameters
    ----------
    seconds: int
      Number of seconds to wait for a second CTRL-C.
    task : str
      A name for the skippable task, to have an explicit script.
    logger : logging.Logger, optional
      The logger to use when printing the messages. The interruption signal is
      notified with ERROR, while the skipping is notified with INFO.
      If not given (default), a brutal print is used.
    """
    if logger:
        err = logger.error
        inf = logger.info
    else:
        err = inf = print
    try:
        yield
    except KeyboardInterrupt:
        err("Received SIGINT. Do it again to stop the script.")
        time.sleep(seconds)
        inf(f"Skipping {task}.")


def save_and_update(
    ds: xr.Dataset,
    pcat: ProjectCatalog,
    path: str | os.PathLike | None = None,
    file_format: str | None = None,
    build_path_kwargs: dict | None = None,
    save_kwargs: dict | None = None,
    update_kwargs: dict | None = None,
):
    """
    Construct the path, save and delete.

    This function can be used after each task of a workflow.

    Parameters
    ----------
    ds: xr.Dataset
        Dataset to save.
    pcat: ProjectCatalog
        Catalog to update after saving the dataset.
    path: str or os.pathlike, optional
        Path where to save the dataset.
        If the string contains variables in curly bracket. They will be filled by catalog attributes.
        If None, the `catutils.build_path` function will be used to create a path.
    file_format: {'nc', 'zarr'}
        Format of the file.
        If None, look for the following in order: build_path_kwargs['format'], a suffix in path, ds.attrs['cat:format'].
        If nothing is found, it will default to zarr.
    build_path_kwargs: dict, optional
        Arguments to pass to `build_path`.
    save_kwargs: dict, optional
        Arguments to pass to `save_to_netcdf` or `save_to_zarr`.
    update_kwargs: dict, optional
        Arguments to pass to `update_from_ds`.
    """
    build_path_kwargs = build_path_kwargs or {}
    save_kwargs = save_kwargs or {}
    update_kwargs = update_kwargs or {}

    # try to guess file format if not given.
    if file_format is None:
        if "format" in build_path_kwargs:
            file_format = build_path_kwargs.get("format")
        elif path is not None and Path(path).suffix:
            file_format = Path(path).suffix.split(".")[-1]
        else:
            file_format = ds.attrs.get("cat:format", "zarr")

    # get path
    if path is not None:
        path = str(path).format(
            **get_cat_attrs(ds, var_as_str=True)
        )  # fill path with attrs
    else:  # if path is not given build it
        build_path_kwargs.setdefault("format", file_format)
        from .catutils import build_path

        path = build_path(ds, **build_path_kwargs)

    # save
    if file_format == "zarr":
        from .io import save_to_zarr

        save_to_zarr(ds, path, **save_kwargs)
    elif file_format == "nc":
        from .io import save_to_netcdf

        save_to_netcdf(ds, path, **save_kwargs)
    else:
        raise ValueError(f"file_format {file_format} is not valid. Use zarr or nc.")

    # update catalog
    pcat.update_from_ds(ds=ds, path=path, **update_kwargs)

    msg = f"File {path} has been saved successfully and the catalog was updated."
    logger.info(msg)


def move_and_delete(
    moving: list[list[str | os.PathLike]],
    pcat: ProjectCatalog,
    deleting: list[str | os.PathLike] | None = None,
    copy: bool = False,
):
    """
    First, move files, then update the catalog with new locations. Finally, delete directories.

    This function can be used at the end of for loop in a workflow to clean temporary files.

    Parameters
    ----------
    moving : list of lists of str or os.PathLike
        list of lists of path of files to move, following the format: [[source 1, destination1], [source 2, destination2],...]
    pcat : ProjectCatalog
        Catalog to update with new destinations
    deleting : list of str or os.PathLike, optional
        List of directories to be deleted, including all contents, and recreated empty. e.g. The working directory of a workflow.
    copy : bool, optional
        If True, copy directories instead of moving them.
    """
    if isinstance(moving, list) and isinstance(moving[0], list):
        for files in moving:
            source, dest = files[0], files[1]
            if Path(source).exists():
                if copy:
                    msg = f"Copying {source} to {dest}."
                    logger.info(msg)
                    copied_files = sh.copytree(source, dest, dirs_exist_ok=True)
                    for f in copied_files:
                        # copied files don't include zarr files
                        if f[-16:] == ".zarr/.zmetadata":
                            with warnings.catch_warnings():
                                # Silence RuntimeWarning about failed guess of backend engines
                                warnings.simplefilter("ignore", category=RuntimeWarning)
                                ds = xr.open_dataset(f[:-11])
                            pcat.update_from_ds(ds=ds, path=f[:-11])
                        if f[-3:] == ".nc":
                            ds = xr.open_dataset(f)
                            pcat.update_from_ds(ds=ds, path=f)
                else:
                    msg = f"Moving {source} to {dest}."
                    logger.info(msg)
                    sh.move(source, dest)
                if Path(dest).suffix in [".zarr", ".nc"]:
                    with warnings.catch_warnings():
                        # Silence RuntimeWarning about failed guess of backend engines
                        warnings.simplefilter("ignore", category=RuntimeWarning)
                        ds = xr.open_dataset(dest)
                    pcat.update_from_ds(ds=ds, path=dest)
            else:
                msg = f"You are trying to move {source}, but it does not exist."
                logger.info(msg)
    else:
        raise ValueError("`moving` should be a list of lists.")

    # erase workdir content if this is the last step
    if isinstance(deleting, list):
        for dir_to_delete in deleting:
            if Path(dir_to_delete).exists() and Path(dir_to_delete).is_dir():
                msg = f"Deleting content inside {dir_to_delete}."
                logger.info(msg)
                sh.rmtree(dir_to_delete)
                Path(dir_to_delete).mkdir()
    elif deleting is None:
        pass
    else:
        raise ValueError("`deleting` should be a list.")



# noqa: D100
#!/usr/bin/env python
#
# xscen documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
import os
import sys
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any

sys.path.insert(0, os.path.abspath(".."))
if os.environ.get("READTHEDOCS") and "ESMFMKFILE" not in os.environ:
    # RTD doesn't activate the env, and esmpy depends on an env var set there
    # We assume the `os` package is in {ENV}/lib/pythonX.X/os.py
    # See conda-forge/esmf-feedstock#91 and readthedocs/readthedocs.org#4067
    os.environ["ESMFMKFILE"] = str(Path(os.__file__).parent.parent / "esmf.mk")

import xscen  # noqa: E402,F401
import xarray  # noqa: E402

xarray.DataArray.__module__ = "xarray"
xarray.Dataset.__module__ = "xarray"

# -- General configuration ---------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "nbsphinx",
    "sphinx_autodoc_typehints",
    "sphinx_codeautolink",
    "sphinx_copybutton",
    "sphinx_mdinclude",
]

# To ensure that underlined fields (e.g. `_field`) are shown in the docs.
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "private-members": False,
    "special-members": False,
}

autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

autosummary_generate = True

nbsphinx_execute = "always"
# To avoid running notebooks on linkcheck and when building PDF.
try:
    skip_notebooks = int(os.getenv("SKIP_NOTEBOOKS"))
except TypeError:
    skip_notebooks = False
if skip_notebooks:
    warnings.warn("SKIP_NOTEBOOKS is set. Not executing notebooks.")
    nbsphinx_execute = "never"
elif (os.getenv("READTHEDOCS_VERSION_NAME") in ["latest", "stable"]) or (
    os.getenv("READTHEDOCS_VERSION_TYPE") in ["tag"]
):
    if Path(__file__).parent.joinpath("notebooks/_data").exists():
        warnings.warn("Notebook artefacts found. Not executing notebooks.")
        nbsphinx_execute = "never"

# if skip_notebooks or os.getenv("READTHEDOCS_VERSION_TYPE") in [
#     "branch",
#     "external",
# ]:
#     warnings.warn("Not executing notebooks.")
#     nbsphinx_execute = "never"

# To avoid having to install these and burst memory limit on ReadTheDocs.
# autodoc_mock_imports = [
#     "cartopy",
#     "clisops",
#     "dask",
#     "h5py",
#     "intake",
#     "intake_esm",
#     "pandas",
#     "rechunker"
#     "xarray",
#     "xclim",
#     "xesmf",
#     "zarr",
# ]

napoleon_numpy_docstring = True
napoleon_use_rtype = False
napoleon_use_param = False
napoleon_use_ivar = True

intersphinx_mapping = {
    "xclim": ("https://xclim.readthedocs.io/en/latest/", None),
    "xsdba": ("https://xsdba.readthedocs.io/en/latest/", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "intake-esm": ("https://intake-esm.readthedocs.io/en/stable/", None),
    "clisops": ("https://clisops.readthedocs.io/en/latest/", None),
    "rechunker": ("https://rechunker.readthedocs.io/en/latest/", None),
    "xesmf": ("https://pangeo-xesmf.readthedocs.io/en/latest/", None),
}

extlinks = {
    "issue": ("https://github.com/Ouranosinc/xscen/issues/%s", "GH/%s"),
    "pull": ("https://github.com/Ouranosinc/xscen/pull/%s", "PR/%s"),
    "user": ("https://github.com/%s", "@%s"),
}

linkcheck_ignore = [
    r"https://github.com/Ouranosinc/xscen/(pull|issue).*",  # too labourious to fully check
    r"https://rmets.onlinelibrary.wiley.com/doi/10.1002/qj.3803",  # Error 403: Forbidden
    r"https://library.wmo.int/idurl/4/56300",  # HTTPconnectionPool error
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a dictionary of suffix: filetype
source_suffix = {'.rst': 'restructuredtext'}

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "xscen"
copyright = f"2022-{datetime.now().year}, Ouranos Inc., Gabriel Rondeau-Genesse, and contributors"
author = "Gabriel Rondeau-Genesse"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = xscen.__version__
# The full version, including alpha/beta/rc tags.
release = xscen.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# Sphinx-intl configuration
locale_dirs = ['locales/']
gettext_compact = False  # optional

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_title = "Ouranos xscen Official Documentation"
html_short_title = "xscen"

html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"logo_only": True, "style_external_links": True}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/_images/xscen-logo.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
if not os.path.exists("_static"):
    os.makedirs("_static")
html_static_path = ["_static"]

# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "xscendoc"

# -- Options for LaTeX output ------------------------------------------

latex_elements: dict[str, Any] = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "xscen.tex",
        "xscen Documentation",
        "Gabriel Rondeau-Genesse",
        "manual",
    ),
]

# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "xscen", "xscen Documentation", [author], 1)]

# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "xscen",
        "xscen Documentation",
        author,
        "xscen",
        "One line description of project.",
        "Miscellaneous",
    ),
]



"""Catalog creation and path building tools."""

import json
import logging
import operator as op
import os
import queue
import string
import threading
import warnings
from collections.abc import Mapping, Sequence
from copy import deepcopy
from fnmatch import fnmatch
from functools import partial, reduce
from itertools import chain, combinations, product
from multiprocessing import Pool
from pathlib import Path
from typing import Any

import cftime
import netCDF4
import numpy as np
import pandas as pd
import parse
import xarray as xr
import yaml
import zarr
from intake_esm import esm_datastore
from pandas import isna

from .catalog import COLUMNS, DataCatalog, generate_id
from .config import parse_config
from .io import get_engine
from .utils import CV, date_parser, ensure_new_xrfreq, get_cat_attrs

logger = logging.getLogger(__name__)


__all__ = [
    "build_path",
    "parse_directory",
    "parse_from_ds",
    "patterns_from_schema",
    "register_parse_type",
]
# ## File finding and path parsing ## #


SUFFIX_TO_FORMAT = {
    ".nc": "nc",
    ".nc4": "nc",
    ".zip": "zarr",
    ".zarr.zip": "zarr",
    ".zarr": "zarr",
}
"""Mapping from file suffix to format.

This is used to populate the "format" esm catalog column from the parsed path.
"""

EXTRA_PARSE_TYPES = {}
"""Extra parse types to add to parse's default.

Add your own types with the :py:func:`register_parse_type` decorator.
"""


def register_parse_type(name: str, regex: str = r"([^\_\/\\]*)", group_count: int = 1):
    r"""Register a new parse type to be available in :py:func:`parse_directory` patterns.

    Function decorated by this will be registered in :py:data:`EXTRA_PARSE_TYPES`.
    The function must take a single string and should return a single string.
    If you return a different type, it may interfere with the other steps of `parse_directory`.

    Parameters
    ----------
    name: str
        The type name. To make use of this type, put "{field:name}" in your pattern.
    regex: str
        A regex string to determine what can be matched by this type.
        The default matches anything but / \ and _, same as the default parse type.
    group_count: int
        The number of regex groups in the previous regex string.
    """

    def _register_parse_type(func):
        EXTRA_PARSE_TYPES[name] = parse.with_pattern(
            regex, regex_group_count=group_count
        )(func)
        return func

    return _register_parse_type


@register_parse_type("no_", regex=r"([^\_\/\\]*)", group_count=1)
def _parse_word(text: str) -> str:
    r"""Parse helper to match strings with anything except / \ or _."""
    return text


@register_parse_type("_", regex=r"([^\/\\]*)", group_count=1)
def _parse_level(text: str) -> str:
    r"""Parse helper to match strings with anything except / or \."""
    return text


# Minimum 4 digits for a date (a single year). Maximum is, in theory, YYYYMMDDHHMMSS so 14.
@register_parse_type(
    "datebounds", regex=r"(([\d]{4,15}(\-[\d]{4,15})?)|fx)", group_count=3
)
def _parse_datebounds(
    text: str,
) -> list[str] | tuple[None, None] | tuple[str, str]:
    """Parse helper to translate date bounds, used in the special DATES field."""
    if "-" in text:
        return text.split("-")
    if text == "fx":
        return None, None
    return text, text


def _find_assets(
    root: str | os.PathLike,
    exts: set[str],
    lengths: set[int],
    dirglob: str | None = None,
    skip_dirs: list[os.PathLike] | None = None,
):
    """Walk recursively over files in a directory, filtering according to a glob pattern, path depth and extensions.

    Parameters
    ----------
    root : str or Pathlike
        Path of the directory to walk through.
    exts : set of strings
        Set of file extensions to look for.
    lengths : set of ints
        Set of path depths to look for.
    dirglob : str, optional
        A glob pattern. If given, only parent folders matching this pattern are walked through.
        This pattern can not include the asset's basename.
    skip_dirs : list of Paths, optional
        A list of directories to skip on the walk.
    """
    skip_dirs = skip_dirs or []
    root = str(Path(root))  # to be sure
    for top, alldirs, files in os.walk(root):
        # Split zarr subdirectories from next iteration
        zarrs = []
        for dr in deepcopy(alldirs):
            fdr = Path(top).joinpath(dr)
            if dr.endswith(".zarr"):
                zarrs.append(dr)
                alldirs.remove(dr)
            if fdr in skip_dirs:
                logger.debug("Skipping %s", fdr)
                alldirs.remove(dr)

        if (
            top != root
            and (os.path.relpath(top, root).count(os.path.sep) + 1) not in lengths
        ):
            continue

        if dirglob is not None and not fnmatch(top, dirglob):
            continue

        if ".zarr" in exts:
            for zr in zarrs:
                yield Path(top).joinpath(zr).as_posix()
        if exts - {".zarr"}:  # There are more exts than
            for file in files:
                if Path(file).suffix in exts:
                    yield Path(top).joinpath(file).as_posix()


def _compile_pattern(pattern: str) -> parse.Parser:
    r"""Compile a parse pattern (if needed) for quicker evaluation.

    The `no_` default format spec is added where no format spec was given.
    The field prefix "?" is converted to "_" so the field name is a valid python variable name.
    """
    if isinstance(pattern, parse.Parser):
        return pattern

    parts = []
    for pre, field, fmt, _ in string.Formatter().parse(pattern):
        if not fmt:
            fmt = "no_"
        if field:
            if field.startswith("?"):
                field = "_" + field[1:]
            if field == "DATES":
                fmt = "datebounds"
            parts.extend([pre, "{", field, ":", fmt, "}"])
        else:
            parts.append(pre)
    return parse.compile("".join(parts), EXTRA_PARSE_TYPES)


def _name_parser(
    path: os.PathLike | str,
    root: os.PathLike | str,
    patterns: list[str | parse.Parser],
    read_from_file: list[str] | dict | None = None,
    attrs_map: dict | None = None,
    xr_open_kwargs: dict | None = None,
) -> dict | None:
    """Extract metadata information from the file path.

    Parameters
    ----------
    path : os.PathLike or str
        Full file path.
    root : os.PathLike or str
        Root directory. Only the part of the path relative to this directory is checked against the patterns.
    patterns : list of str or parse.Parser
        List of patterns to try in `parse.parse`. See :py:func:`parse_directory` for the pattern specification.
    read_from_file : list of string or dict, optional
        If not None, passed directly to :py:func:`parse_from_ds` as `names`.
        If None (default), only the path is parsed, the file is not opened.
    attrs_map : dict, optional
        If `read_from_file` is not None, passed directly to :py:func:`parse_from_ds`.
    xr_open_kwargs : dict, optional
        If `read_from_file` is not None, passed directly to :py:func:`parse_from_ds`.

    Returns
    -------
    dict or None
        The metadata fields parsed from the path using the first matching pattern.
        If no pattern matched, None is returned.

    See Also
    --------
    parse.parse
    parse_directory
    parse_from_ds
    """
    abs_path = Path(path)
    path = abs_path.relative_to(Path(root))
    xr_open_kwargs = xr_open_kwargs or {}

    d = {}
    for pattern in map(_compile_pattern, patterns):
        res = pattern.parse(str(path))
        if res:
            d = res.named
            break
    else:
        return None

    d["path"] = abs_path
    d["format"] = SUFFIX_TO_FORMAT.get(path.suffix, path.suffix[1:])

    if "DATES" in d:
        d["date_start"], d["date_end"] = d.pop("DATES")

    if read_from_file:
        fromfile = parse_from_ds(
            abs_path, names=read_from_file, attrs_map=attrs_map, **xr_open_kwargs
        )
        d.update(fromfile)

    # files with a single year/month
    if "date_end" not in d and "date_start" in d:
        d["date_end"] = d["date_start"]

    # strip to clean off lost spaces and line jumps
    # do not include wildcarded fields (? was transformed to _ in _compile_pattern)
    return {
        k: v.strip() if isinstance(v, str) else v
        for k, v in d.items()
        if not k.startswith("_")
    }


def _parse_dir(  # noqa: C901
    root: os.PathLike | str,
    patterns: list[str],
    dirglob: str | None = None,
    skip_dirs: list[os.PathLike] | None = None,
    checks: list[str] | None = None,
    read_from_file: list[str] | dict | None = None,
    attrs_map: dict | None = None,
    xr_open_kwargs: dict | None = None,
    progress: bool = False,
):
    """Iterate and parses files in a directory, filtering according to basic pattern properties and optional checks.

    Parameters
    ----------
    root : os.PathLike or str
        Path to walk through.
    patterns : list of strings or compiled parsers
        Patterns that the files will be checked against.
        The extensions of the patterns are extracted and only paths with these are returned.
        Also, the depths of the patterns are calculated and only paths of this depth under the root are returned.
    dirglob : str
        A glob pattern. If given, only parent folders matching this pattern are walked through.
        This pattern can not include the asset's basename.
    skip_dirs : list of strings or Paths, optional
        A list of directories to skip in the walk.
    checks: list of strings, optional
        A list of checks to perform, available values are:
        - "readable" : Check that the file is readable by the current user.
        - "writable" : Check that the file is writable by the current user.
        - "ncvalid" : For netCDF, check that it is valid (openable with netCDF4).
        All checks will slow down the parsing.
    read_from_file : list of string or dict, optional
        If not None, passed directly to :py:func:`parse_from_ds` as `names`.
        If None (default), only the path is parsed, the file is not opened.
    attrs_map : dict, optional
        If `read_from_file` is not None, passed directly to :py:func:`parse_from_ds`.
    xr_open_kwargs : dict, optional
        If `read_from_file` is not None, passed directly to :py:func:`parse_from_ds`.
    progress : bool
        If True, the number of found files is printed to stdout.

    Return
    ------
    List of dictionaries
        Metadata parsed from each found asset.
    """
    lengths = {patt.count(os.path.sep) for patt in patterns}
    exts = {Path(patt).suffix for patt in patterns}
    comp_patterns = list(map(_compile_pattern, patterns))
    checks = checks or []
    parsed = []

    root = Path(root)
    if any([(skd in root.parents) or (skd == root) for skd in (skip_dirs or [])]):
        logger.debug("Skipping %s", root)
        return parsed

    # Multithread, communicating via FIFO queues.
    # This thread walks the directory
    # Another thread runs the checks
    # Another thread parses the path and file.
    # In theory, for a local disk, walking a directory cannot be parallelized. This is not as true for network-mounted drives.
    # Thus, we parallelize the parsing steps.
    # If the name-parsing step becomes blocking, we could try to increase the number of threads (but netCDF4 can't multithread...)
    # Usually, the walking is the bottleneck.
    q_found = queue.Queue()
    q_checked = queue.Queue()

    def check_worker():
        # Worker that processes the checks.
        while True:
            path = q_found.get()
            valid = True
            if "readable" in checks and not os.access(path, os.R_OK):
                valid = False
            if "writable" in checks and not os.access(path, os.W_OK):
                valid = False
            if "ncvalid" in checks:
                try:  # Simple check that the file is openable
                    if get_engine(path) == "netcdf4":
                        # if get_engine is "h5netcdf", it means h5py was able to recognize it.
                        # TODO: testing for zarr validity is not implemented
                        with netCDF4.Dataset(path):
                            pass
                # FIXME: This is a catch-all, we should catch the specific exception raised by netCDF4.
                except Exception:  # noqa: BLE001
                    valid = False
            if valid:
                q_checked.put(path)
            q_found.task_done()

    def parse_worker():
        # Worker that parses the paths
        while True:
            path = q_checked.get()
            try:
                d = _name_parser(
                    path,
                    root,
                    comp_patterns,
                    read_from_file=read_from_file,
                    attrs_map=attrs_map,
                    xr_open_kwargs=xr_open_kwargs,
                )
            # FIXME: This is not specific enough, we should catch the specific exception raised by _name_parser.
            except Exception as err:  # noqa: BLE001
                msg = f"Parsing file {path} failed with {err}."
                logger.error(msg)
            else:
                if d is not None:
                    parsed.append(d)
                    n = len(parsed)
                    # Print number of files but on round numbers to limit the calls to stdout for large collections
                    if progress and all(
                        [(n < N or (n % N == 0)) for N in [10, 100, 1000]]
                    ):
                        print(f"Found {n:7d} files", end="\r")
                else:
                    msg = f"File {path} didn't match any pattern."
                    logger.debug(msg)
            q_checked.task_done()

    CW = threading.Thread(target=check_worker, daemon=True)
    CW.start()

    PW = threading.Thread(target=parse_worker, daemon=True)
    PW.start()

    # Skip the checks if none are requested (save some overhead)
    q = q_found if checks else q_checked
    for path in _find_assets(Path(root), exts, lengths, dirglob, skip_dirs):
        q.put(path)

    q_found.join()
    q_checked.join()
    return parsed


def _get_new_item(name, newval, repval, oldval, fromcol, is_list):
    if is_list:
        if name == fromcol:  # We replace only the repval element of the list
            return tuple(newval if v == repval else v for v in oldval)
        # We must return a tuple, replace the whole list with a single element.
        return (newval,)
    return newval  # Simple replacement


def _replace_in_row(oldrow: pd.Series, replacements: dict):
    """Replace values in Series (row) according to replacements mapping.

    Replacements can be simple mappings, but also mapping to other fields.
    List-like fields are handled.
    """
    row = oldrow.copy()
    list_cols = [col for col in oldrow.index if isinstance(oldrow[col], tuple | list)]
    for col, reps in replacements.items():
        if col not in row:
            continue
        for repval, new in reps.items():
            # Either the field is a list containing the value to replace, or it is the value to replace.
            if (col in list_cols and repval in row[col]) or repval == row[col]:
                if isinstance(new, dict):  # Replacement is for multiple columns
                    for name, newval in new.items():
                        row[name] = _get_new_item(
                            name, newval, repval, row[col], col, name in list_cols
                        )
                else:
                    row[col] = _get_new_item(
                        col, new, repval, row[col], col, col in list_cols
                    )
    # Special case for "variable" where we remove Nones.
    if "variable" in row and "variable" in list_cols and None in row["variable"]:
        row["variable"] = tuple(v for v in row["variable"] if v is not None)
    return row


def _parse_first_ds(
    grp: pd.DataFrame, cols: list[str], attrs_map: dict, xr_open_kwargs: dict
):
    """Parse attributes from one file per group, apply them to the whole group."""
    fromfile = parse_from_ds(grp.path.iloc[0], cols, attrs_map, **xr_open_kwargs)

    msg = f"Got {len(fromfile)} fields, applying to {len(grp)} entries."
    logger.info(msg)
    out = grp.copy()
    for col, val in fromfile.items():
        for i in grp.index:  # If val is an iterable we can't use loc.
            out.at[i, col] = val
    return out


@parse_config
def parse_directory(  # noqa: C901
    directories: str | list[str | os.PathLike],
    patterns: list[str],
    *,
    id_columns: list[str] | None = None,
    read_from_file: (
        bool
        | Sequence[str]
        | tuple[Sequence[str], Sequence[str]]
        | Sequence[tuple[Sequence[str], Sequence[str]]]
    ) = False,
    homogenous_info: dict | None = None,
    cvs: str | os.PathLike | dict | None = None,
    dirglob: str | None = None,
    skip_dirs: list[str | os.PathLike] | None = None,
    xr_open_kwargs: Mapping[str, Any] | None = None,
    only_official_columns: bool = True,
    progress: bool = False,
    parallel_dirs: bool | int = False,
    file_checks: list[str] | None = None,
) -> pd.DataFrame:
    r"""Parse files in a directory and return them as a pd.DataFrame.

    Parameters
    ----------
    directories : list of os.PathLike or list of str
        List of directories to parse. The parse is recursive.
    patterns : list of str
        List of possible patterns to be used by :py:func:`parse.parse` to decode the file names. See Notes below.
    id_columns : list of str, optional
        List of column names on which to base the dataset definition. Empty columns will be skipped.
        If None (default), it uses :py:data:`ID_COLUMNS`.
    read_from_file : boolean or set of strings or tuple of 2 sets of strings or list of tuples
        If True, if some fields were not parsed from their path, files are opened and
        missing fields are parsed from their metadata, if found.
        If a sequence of column names, only those fields are parsed from the file, if missing.
        If False (default), files are never opened.
        If a tuple of 2 lists of strings, only the first file of groups defined by the first list of
        columns is read and the second list of columns is parsed from the file and applied to the whole group.
        For example, `(["source"],["institution", "activity"])` will find a group with all the files
        that have the same source, open only one of the files to read the institution
        and activity, and write this information in the catalog for all files of the group.
        It can also be a list of those tuples.
    homogenous_info : dict, optional
        Using the {column_name: description} format, information to apply to all files.
        These are applied before the `cvs`.
    cvs: str or os.PathLike or dict, optional
        Dictionary with mapping from parsed term to preferred terms (Controlled VocabularieS) for each column.
        May have an additional "attributes" entry which maps from attribute names in the files to
        official column names. The attribute translation is done before the rest.
        In the "variable" entry, if a name is mapped to None (null), that variable will not be listed in the catalog.
        A term can map to another mapping from field name to values, so that a value on one column triggers the filling of other columns.
        In the latter case, that other column must exist beforehand, whether it was in the pattern or in the homogenous_info.
    dirglob : str, optional
        A glob pattern for path matching to accelerate the parsing of a directory tree if only a subtree is needed.
        Only folders matching the pattern are parsed to find datasets.
    skip_dirs : list of str or Paths, optional
        A list of folders that will be removed from the search, should be absolute.
    xr_open_kwargs: dict
        If needed, arguments to send xr.open_dataset() when opening the file to read the attributes.
    only_official_columns: bool
        If True (default), this ensures the final catalog only has the columns defined in :py:data:`xscen.catalog.COLUMNS`.
        Other fields in the patterns will raise an error.
        If False, the columns are those used in the patterns and the homogeneous info.
        In that case, the column order is not determined.
        Path, format, and id are always present in the output.
    progress : bool
        If True, a counter is shown in stdout when finding files on disk. Does nothing if `parallel_dirs` is not False.
    parallel_dirs: bool or int
        If True, each directory is searched in parallel. If an int, it is the number of parallel searches.
        This should only be significantly useful if the directories are on different disks.
    file_checks: list of str, optional
        A list of file checks to run on the parsed files. Available values are:
        - "readable" : Check that the file is readable by the current user.
        - "writable" : Check that the file is writable by the current user.
        - "ncvalid" : For netCDF, check that it is valid (openable with netCDF4).
        Any check will slow down the parsing.

    Notes
    -----
    - Official columns names are controlled and ordered by :py:data:`COLUMNS`:
        ["id", "type", "processing_level", "mip_era", "activity", "driving_institution", "driving_model", "institution",
         "source", "bias_adjust_institution", "bias_adjust_project","experiment", "member",
         "xrfreq", "frequency", "variable", "domain", "date_start", "date_end", "version"]
    - Not all column names have to be present, but "xrfreq" (obtainable through "frequency"), "variable",
        "date_start" and "processing_level" are necessary for a workable catalog.
    - 'patterns' should highlight the columns with braces.
        This acts like the reverse operation of `format()`. It is a template string with `{field name:type}` elements.
        The default "type" will match alphanumeric parts of the path, excluding the "_", "/" and "\" characters.
        The "_" type will allow underscores.
        Field names prefixed by "?" will not be included in the output.
        See the documentation of :py:mod:`parse` for more type options.
        You can also add your own types using the :py:func:`register_parse_type` decorator.

        The "DATES" field is special as it will only match dates, either as a single date (YYYY, YYYYMM, YYYYMMDD)
        assigned to "{date_start}" (with "date_end" automatically inferred) or two dates of the same format as "{date_start}-{date_end}".

        Example: `"{source}/{?ignored project name}_{?:_}_{DATES}.nc"`
        Here, "source" will be the full folder name and it can't include underscores.
        The first section of the filename will be excluded from the output, it was given a name (ignore project name) to make the pattern readable.
        The last section of the filenames ("dates") will yield a "date_start" / "date_end" couple.
        All other sections in the middle will be ignored, as they match "{?:_}".

    Returns
    -------
    pd.DataFrame
        Parsed directory files
    """
    if isinstance(directories, str | Path):
        directories = [directories]
    homogenous_info = homogenous_info or {}
    xr_open_kwargs = xr_open_kwargs or {}
    if only_official_columns:
        columns = set(COLUMNS) - homogenous_info.keys()
        pattern_fields = {
            f
            for f in set.union(
                *(set(patt.named_fields) for patt in map(_compile_pattern, patterns))
            )
            if not f.startswith("_")
        } - {"DATES"}
        unrecognized = pattern_fields - set(COLUMNS)
        if unrecognized:
            raise ValueError(
                f"Patterns include fields which are not recognized by xscen : {unrecognized}. "
                "If this is wanted, pass only_official_columns=False to remove the check."
            )

    read_file_groups = False  # Whether to read file per group or not.
    if not isinstance(read_from_file, bool) and not isinstance(read_from_file[0], str):
        # A tuple of 2 lists
        read_file_groups = True
        if isinstance(read_from_file[0][0], str):
            # only one grouping
            read_from_file = [read_from_file]
    elif read_from_file is True:
        # True but not a list of strings
        read_from_file = columns

    if cvs is not None:
        if not isinstance(cvs, dict):
            with Path(cvs).open(encoding="utf-8") as f:
                cvs = yaml.safe_load(f)
        attrs_map = cvs.pop("attributes", {})
    else:
        attrs_map = {}

    parse_kwargs = dict(
        patterns=patterns,
        dirglob=dirglob,
        skip_dirs=[Path(d) for d in (skip_dirs or [])],
        read_from_file=read_from_file if not read_file_groups else None,
        attrs_map=attrs_map,
        xr_open_kwargs=xr_open_kwargs,
        checks=file_checks,
    )

    if parallel_dirs is True:
        parallel_dirs = len(directories)

    parsed = []
    if parallel_dirs > 1:
        with Pool(processes=parallel_dirs) as pool:
            results = []
            for directory in directories:
                results.append(pool.apply_async(_parse_dir, (directory,), parse_kwargs))
            for res in results:
                parsed.extend(res.get())
    else:
        for directory in directories:
            parsed.extend(_parse_dir(directory, progress=progress, **parse_kwargs))

    if not parsed:
        raise ValueError("No files found.")
    else:
        if progress:
            print()  # This is because of the \r outputted in the _parse_dir call.
        msg = f"Found and parsed {len(parsed)} files."
        logger.info(msg)

    # Path has become NaN when some paths didn't fit any passed pattern
    df = pd.DataFrame(parsed).dropna(axis=0, subset=["path"])

    if only_official_columns:  # Add the missing official columns
        for col in set(COLUMNS) - set(df.columns):
            df[col] = None

    if read_file_groups:  # Read fields from file, but only one per group.
        for group_cols, parse_cols in read_from_file:
            df = (  # column indexing avoids a deprecation where cols in group_cols are not sent to apply
                df.groupby(group_cols)[df.columns]
                .apply(
                    _parse_first_ds,
                    cols=parse_cols,
                    attrs_map=attrs_map,
                    xr_open_kwargs=xr_open_kwargs,
                )
                .reset_index(drop=True)
            )

    # Everything below could be wrapped in a function to be applied to each row maybe allowing some basic parallelization with dask (or else).
    # Add homogeous info
    for key, val in homogenous_info.items():
        df[key] = val

    # Replace entries by definitions found in CV
    if cvs:
        df = df.apply(_replace_in_row, axis=1, replacements=cvs)

    # Fix potential legacy xrfreq
    if "xrfreq" in df.columns:
        df["xrfreq"] = df["xrfreq"].apply(ensure_new_xrfreq)

    # translate xrfreq into frequencies and vice-versa
    if {"xrfreq", "frequency"}.issubset(df.columns):
        df.fillna(
            {"xrfreq": df["frequency"].apply(CV.frequency_to_xrfreq, default=pd.NA)},
            inplace=True,
        )
        df.fillna(
            {"frequency": df["xrfreq"].apply(CV.xrfreq_to_frequency, default=pd.NA)},
            inplace=True,
        )

    # Parse dates
    # If we don't do the to_numpy(na_value=np.datetime64('')).astype('<M8[ms]') trick,
    # the dtype will be "object" if any of the dates are out-of-bounds.
    # `na_values=np.datetime64('')` is needed because pandas' NaT does not translate to numpy's NaT, but to float.
    if "date_start" in df.columns:
        df["date_start"] = (
            df["date_start"]
            .apply(date_parser)
            .to_numpy(na_value=np.datetime64(""))
            .astype("<M8[ms]")
        )
    if "date_end" in df.columns:
        df["date_end"] = (
            df["date_end"]
            .apply(date_parser, end_of_period=True)
            .to_numpy(na_value=np.datetime64(""))
            .astype("<M8[ms]")
        )
    # Checks
    if {"date_start", "date_end", "xrfreq", "frequency"}.issubset(df.columns):
        # All NaN dates correspond to a fx frequency.
        invalid = df.date_start.isnull() & df.date_end.isnull() & (df.xrfreq != "fx")
        n = invalid.sum()
        if n > 0:
            warnings.warn(
                f"{n} invalid entries where the start and end dates are Null but the frequency is not 'fx'."
            )
            msg = f"Paths: {df.path[invalid].values}"
            logger.debug(msg)
            df = df[~invalid]
        # Exact opposite
        invalid = df.date_start.notnull() & df.date_end.notnull() & (df.xrfreq == "fx")
        n = invalid.sum()
        if n > 0:
            warnings.warn(
                f"{n} invalid entries where the start and end dates are given but the frequency is 'fx'."
            )
            msg = f"Paths: {df.path[invalid].values}"
            logger.debug(msg)
            df = df[~invalid]

    # Create id from user specifications
    df["id"] = generate_id(df, id_columns)

    # TODO: ensure variable is a tuple ?

    # ensure path is a string
    df["path"] = df.path.apply(str)

    # Sort columns and return
    if only_official_columns:
        return df.loc[:, COLUMNS]
    return df


def parse_from_ds(  # noqa: C901
    obj: str | os.PathLike | xr.Dataset,
    names: Sequence[str],
    attrs_map: Mapping[str, str] | None = None,
    **xrkwargs,
):
    """Parse a list of catalog fields from the file/dataset itself.

    If passed a path, this opens the file.

    Infers the variable from the variables.
    Infers xrfreq, frequency, date_start and date_end from the time coordinate if present.
    Infers other attributes from the coordinates or the global attributes. Attributes names
    can be translated using the `attrs_map` mapping (from file attribute name to name in `names`).

    If the obj is the path to a Zarr dataset and none of "frequency", "xrfreq", "date_start" or "date_end"
    are requested, :py:func:`parse_from_zarr` is used instead of opening the file.

    Parameters
    ----------
    obj: str or os.PathLike or xr.Dataset
        Dataset to parse.
    names: sequence of str
        List of attributes to be parsed from the dataset.
    attrs_map: dict, optional
        In the case of non-standard names in the file, this can be used to match entries in the files to specific 'names' in the requested list.
    xrkwargs:
        Arguments to be passed to open_dataset().
    """
    get_time = bool(
        {"frequency", "xrfreq", "date_start", "date_end"}.intersection(names)
    )
    if not isinstance(obj, xr.Dataset):
        obj = Path(obj)

    if isinstance(obj, Path) and obj.suffixes[-1] == ".zarr":
        msg = f"Parsing attributes from Zarr {obj}."
        logger.info(msg)
        ds_attrs, variables, time = _parse_from_zarr(
            obj, get_vars="variable" in names, get_time=get_time
        )
    elif isinstance(obj, Path) and obj.suffixes[-1] == ".nc":
        msg = f"Parsing attributes with netCDF4 from {obj}."
        logger.info(msg)
        ds_attrs, variables, time = _parse_from_nc(
            obj, get_vars="variable" in names, get_time=get_time
        )
    else:
        if isinstance(obj, Path):
            msg = f"Parsing attributes with xarray from {obj}."
            logger.info(msg)
            obj = xr.open_dataset(obj, engine=get_engine(obj), **xrkwargs)
        ds_attrs = obj.attrs
        time = obj.indexes["time"] if "time" in obj else None
        variables = set(obj.data_vars.keys()).difference(
            [v for v in obj.data_vars if len(obj[v].dims) == 0]
        )

    rev_attrs_map = {v: k for k, v in (attrs_map or {}).items()}
    attrs = {}

    for name in names:
        if name == "variable":
            attrs["variable"] = tuple(sorted(variables))
        elif name in ("frequency", "xrfreq") and time is not None and time.size > 3:
            # round to the minute to catch floating point imprecision
            freq = xr.infer_freq(time.round("min"))
            if freq:
                if "xrfreq" in names:
                    attrs["xrfreq"] = freq
                if "frequency" in names:
                    attrs["frequency"] = CV.xrfreq_to_frequency(freq)
            else:
                warnings.warn(
                    f"Couldn't infer frequency of dataset {obj if not isinstance(obj, xr.Dataset) else ''}"
                )
        elif name in ("frequency", "xrfreq") and time is None:
            attrs[name] = "fx"
        elif name == "date_start" and time is not None:
            attrs["date_start"] = time[0]
        elif name == "date_end" and time is not None:
            attrs["date_end"] = time[-1]
        elif name in rev_attrs_map and rev_attrs_map[name] in ds_attrs:
            attrs[name] = ds_attrs[rev_attrs_map[name]].strip()
        elif name in ds_attrs:
            attrs[name] = ds_attrs[name].strip()

    msg = f"Got fields {attrs.keys()} from file."
    logger.debug(msg)
    return attrs


def _parse_from_zarr(
    path: os.PathLike | str, get_vars: bool = True, get_time: bool = True
):
    """Obtain the list of variables, the time coordinate and the list of global attributes from a zarr dataset.

    Vars and attrs from reading the JSON files directly, time by reading the data with zarr.

    Variables are those
    - where .zattrs/_ARRAY_DIMENSIONS is not empty
    - where .zattrs/_ARRAY_DIMENSIONS does not contain the variable name
    - who do not appear in any "coordinates" attribute.

    Parameters
    ----------
    path: os.PathLike or str
        Path to the zarr dataset.
    get_vars: bool
        If True, return the list of variables.
    get_time: bool
        If True, return the time coordinate.
    """
    path = Path(path)

    if (path / ".zattrs").is_file():
        with (path / ".zattrs").open() as f:
            ds_attrs = json.load(f)
    else:
        ds_attrs = {}

    variables = []
    if get_vars:
        coords = []
        for varpath in path.iterdir():
            if varpath.is_dir() and (varpath / ".zattrs").is_file():
                with (varpath / ".zattrs").open() as f:
                    var_attrs = json.load(f)
                if (
                    varpath.name in var_attrs["_ARRAY_DIMENSIONS"]
                    or len(var_attrs["_ARRAY_DIMENSIONS"]) == 0
                ):
                    coords.append(varpath.name)
                if "coordinates" in var_attrs:
                    coords.extend(
                        list(map(str.strip, var_attrs["coordinates"].split(" ")))
                    )
        variables = [
            varpath.name
            for varpath in path.iterdir()
            if varpath.name not in coords and varpath.is_dir()
        ]
    time = None
    if get_time and (path / "time").is_dir():
        ds = zarr.open(path)
        time = xr.CFTimeIndex(
            cftime.num2date(
                ds.time[:],
                calendar=ds.time.attrs["calendar"],
                units=ds.time.attrs["units"],
            )
        )
    return ds_attrs, variables, time


def _parse_from_nc(
    path: os.PathLike | str, get_vars: bool = True, get_time: bool = True
):
    """Obtain the list of variables, the time coordinate, and the list of global attributes from a netCDF dataset, using netCDF4.

    Parameters
    ----------
    path: os.PathLike or str
        Path to the netCDF dataset.
    get_vars: bool
        If True, return the list of variables.
    get_time: bool
        If True, return the time coordinate.
    """
    ds = netCDF4.Dataset(str(Path(path)))
    ds_attrs = {k: ds.getncattr(k) for k in ds.ncattrs()}

    variables = []
    if get_vars:
        coords = []
        for name, var in ds.variables.items():
            if "coordinates" in var.ncattrs():
                coords.extend(
                    list(map(str.strip, var.getncattr("coordinates").split(" ")))
                )
            if len(var.dimensions) == 0 or name in var.dimensions:
                coords.append(name)
        variables = [var for var in ds.variables.keys() if var not in coords]

    time = None
    if get_time and "time" in ds.variables:
        time = xr.CFTimeIndex(
            cftime.num2date(
                ds["time"][:], calendar=ds["time"].calendar, units=ds["time"].units
            ).data
        )
    ds.close()
    return ds_attrs, variables, time


# ## Path building ## #
def _schema_option(option: dict, facets: dict):
    """Parse an option element of the facet schema tree."""
    facet_value = facets.get(option["facet"])
    if "value" in option:
        if isinstance(option["value"], str):
            answer = facet_value == option["value"]
        else:  # A list
            answer = facet_value in option["value"]
    else:
        answer = not isna(facet_value)
    return answer


def _schema_level(schema: dict | list[str] | str, facets: dict):
    if isinstance(schema, str):
        if schema.startswith("(") and schema.endswith(")"):
            optional = True
            schema = schema[1:-1]
        else:
            optional = False
        if schema == "DATES":
            return _schema_dates(facets, optional=optional)

        # A single facet:
        if isna(facets.get(schema)):
            if optional:
                return None
            raise ValueError(
                f"Facet {schema} is needed but None-like or missing in the data."
            )
        return facets[schema]
    if isinstance(schema, list):
        parts = []
        for element in schema:
            part = _schema_level(element, facets)
            if not isna(part):
                parts.append(part)
        return "_".join(parts)
    if isinstance(schema, dict) and "text" in schema:
        return schema["text"]
    raise ValueError(f"Invalid schema : {schema}")


def _schema_dates(facets: dict, optional: bool = False):
    if facets.get("xrfreq") == "fx":
        return "fx"

    if any([facets.get(f) is None for f in ["date_start", "date_end", "xrfreq"]]):
        if optional:
            return None
        raise ValueError(
            "Facets date_start, date_end and xrfreq are needed, but at least one is missing or None-like in the data."
        )

    start = date_parser(facets["date_start"])
    end = date_parser(facets["date_end"])
    freq = pd.Timedelta(CV.xrfreq_to_timedelta(facets["xrfreq"]))

    # Full years : Starts on Jan 1st and is either annual or ends on Dec 31st (accepting Dec 30 for 360 cals)
    if (
        start.month == 1
        and start.day == 1
        and (
            freq >= pd.Timedelta(CV.xrfreq_to_timedelta("YS"))
            or (end.month == 12 and end.day > 29)
        )
    ):
        if start.year == end.year:
            return f"{start:%4Y}"
        return f"{start:%4Y}-{end:%4Y}"
    # Full months : Starts on the 1st and is either monthly or ends on the last day
    if start.day == 1 and (
        freq >= pd.Timedelta(CV.xrfreq_to_timedelta("M")) or end.day > 27
    ):
        # Full months
        if (start.year, start.month) == (end.year, end.month):
            return f"{start:%4Y%m}"
        return f"{start:%4Y%m}-{end:%4Y%m}"
    # The full range
    return f"{start:%4Y%m%d}-{end:%4Y%m%d}"


def _schema_filename(schema: list, facets: dict) -> str:
    return "_".join(
        [
            facets.get(element) if element != "DATES" else _schema_dates(facets)
            for element in schema
            if element == "DATES" or not isna(facets.get(element))
        ]
    )


def _schema_folders(schema: list, facets: dict) -> list:
    folder_tree = list()
    for level in schema:
        part = _schema_level(level, facets)
        if not isna(part):
            folder_tree.append(part)
    return folder_tree


def _get_needed_fields(schema: dict):
    """Return the list of facets that is needed for a given schema."""
    needed = set()
    for level in schema["folders"]:
        if isinstance(level, str):
            if not (level.startswith("(") and level.endswith(")")):
                needed.add(level)
        elif isinstance(level, list):
            for lvl in level:
                if not (lvl.startswith("(") and lvl.endswith(")")):
                    needed.add(lvl)
        elif not (isinstance(level, dict) and list(level.keys()) == ["text"]):
            raise ValueError(
                f"Invalid schema with unknown {level} of type {type(level)}."
            )
    return needed


def _read_schemas(schemas):
    if isinstance(schemas, dict) and {"folders", "filename"}.issubset(schemas.keys()):
        # Single schema
        # Remove any conditions (or insert empty one)
        schemas["with"] = []
        schemas = {"unnamed_schema": schemas}
    elif not isinstance(schemas, dict):
        if schemas is None:
            schemas = Path(__file__).parent / "data" / "file_schema.yml"
        with Path(schemas).open(encoding="utf-8") as f:
            schemas = yaml.safe_load(f)
    for name, schema in schemas.items():
        missing_fields = {"with", "folders", "filename"} - set(schema.keys())
        if missing_fields:
            raise ValueError(
                f"Invalid schema specification. Missing fields {missing_fields} in schema {name}."
            )
    return schemas


def _build_path(
    data: dict | xr.Dataset | xr.DataArray | pd.Series,
    schemas: dict,
    root: str | os.PathLike,
    get_type: bool = False,
    **extra_facets,
) -> Path | tuple[Path, str]:
    # Get all known metadata
    if isinstance(data, xr.Dataset | xr.DataArray):
        facets = (
            # Get non-attribute metadata
            parse_from_ds(
                data, ["frequency", "xrfreq", "date_start", "date_end", "variable"]
            )
            | data.attrs
            | get_cat_attrs(data)
        )
    elif isinstance(data, pd.Series):
        facets = dict(data)
    else:
        raise NotImplementedError(f"Cannot build path with object of type {type(data)}")

    facets = facets | extra_facets

    # Scalar-ize variable if needed.
    if (
        "variable" in facets
        and not isinstance(facets["variable"], str)
        and len(facets["variable"]) == 1
    ):
        facets["variable"] = facets["variable"][0]

    # Find the first fitting schema
    for name, schema in schemas.items():
        if not schema["with"]:
            match = True
        else:
            match = reduce(
                op.and_, map(partial(_schema_option, facets=facets), schema["with"])
            )
        if match:
            # Checks
            needed_fields = _get_needed_fields(schema)
            if missing_fields := needed_fields - set(facets.keys()):
                raise ValueError(
                    f"Missing facets {missing_fields} are needed to build the path according to selected schema {name}."
                )
            if "variable" in needed_fields and not isinstance(facets["variable"], str):
                raise ValueError(
                    f"Selected schema {name} is meant to be used with single-variable datasets. Got multiple: {facets['variable']}. "
                    "You can override the facet by passing `variable='varname'` directly."
                )
            out = Path(*_schema_folders(schema["folders"], facets))
            out = out / _schema_filename(schema["filename"], facets)
            if root is not None:
                out = Path(root) / out
            if "format" in facets:  # Add extension
                # Can't use `with_suffix` in case there are dots in the name
                out = out.parent / f"{out.name}.{facets['format']}"
            if get_type:
                return out, name
            return out

    raise ValueError(f"This file doesn't match any schema. Facets:\n{facets}")


@parse_config
def build_path(
    data: dict | xr.Dataset | xr.DataArray | pd.Series | DataCatalog | pd.DataFrame,
    schemas: str | os.PathLike | dict | None = None,
    root: str | os.PathLike | None = None,
    **extra_facets,
) -> Path | DataCatalog | pd.DataFrame:
    r"""Parse the schema from a configuration and construct path using a dictionary of facets.

    Parameters
    ----------
    data : dict or xr.Dataset or xr.DataArray or pd.Series or DataCatalog or pd.DataFrame
        Dict of facets. Or xarray object to read the facets from. In the latter case, variable and time-dependent
        facets are read with :py:func:`parse_from_ds` and supplemented with all the object's attribute,
        giving priority to the "official" xscen attributes (prefixed with `cat:`, see :py:func:`xscen.utils.get_cat_attrs`).
        Can also be a catalog or a DataFrame, in which a "new_path" column is generated for each item.
    schemas : Path or dict, optional
        Path to YAML schematic of database schema. If None, will use a default schema.
        See the comments in the `xscen/data/file_schema.yml` file for more details on its construction.
        A dict of dict schemas can be given (same as reading the yaml).
        Or a single schema dict (single element of the yaml).
    root : str or Path, optional
        If given, the generated path(s) is given under this root one.
    \*\*extra_facets
        Extra facets to supplement or override metadadata missing from the first input.

    Returns
    -------
    Path or catalog
        Constructed path. If "format" is absent from the facets, it has no suffix.
        If `data` was a catalog, a copy with a "new_path" column is returned.
        Another "new_path_type" column is also added if `schemas` was a collection of schemas (like the default).

    Examples
    --------
    To rename a full catalog, the simplest way is to do:

    >>> import xscen as xs
    >>> import shutil as sh
    >>> new_cat = xs.catutils.build_path(old_cat)
    >>> for i, row in new_cat.iterrows():
    ...     sh.move(row.path, row.new_path)
    ...
    """
    if root:
        root = Path(root)
    schemas = _read_schemas(schemas)
    if isinstance(data, esm_datastore | pd.DataFrame):
        if isinstance(data, esm_datastore):
            df = data.df
        else:
            df = data

        df = df.copy()

        paths = df.apply(
            _build_path,
            axis=1,
            result_type="expand",
            schemas=schemas,
            root=root,
            get_type=True,
            **extra_facets,
        )
        df["new_path"] = paths[0].apply(str)
        if len(schemas) > 1:
            df["new_path_type"] = paths[1]
        return df
    return _build_path(data, schemas=schemas, root=root, get_type=False, **extra_facets)


def _as_template(a):
    return "{" + a + "}"


def partial_format(template, **fmtargs):
    """Format a template only partially, leaving un-formatted templates intact."""

    class PartialFormatDict(dict):
        def __missing__(self, key):
            return _as_template(key)

    return template.format_map(PartialFormatDict(**fmtargs))


def patterns_from_schema(schema: str | dict, exts: Sequence[str] | None = None):
    """Generate all valid patterns for a given schema.

    Generated patterns are meant for use with :py:func:`parse_directory`.
    This hardcodes the rule that facet can never contain a underscore ("_") except "variable".
    File names are not strict except for the date bounds element which must be at the end if present.

    Parameters
    ----------
    schema: dict or str
        A dict with keys "with" (optional), "folders" and "filename", constructed as described
        in the `xscen/data/file_schema.yml` file.
        Or the name of a pattern group from that file.
    exts: sequence of strings, optional
        A list of file extensions to consider, with the leading dot.
        Defaults to ``[".nc", ".zarr", ".zarr.zip"]``.

    Returns
    -------
    list of patterns compatible with :py:func:`parse_directory`.
    """
    if isinstance(schema, str):
        schemas = Path(__file__).parent / "data" / "file_schema.yml"
        with schemas.open(encoding="utf-8") as f:
            schema = yaml.safe_load(f)[schema]

    # # Base folder patterns

    # Index of optional folder parts
    opt_idx = [
        i
        for i, k in enumerate(schema["folders"])
        if isinstance(k, str) and k.startswith("(")
    ]

    raw_folders = []
    for skip in chain.from_iterable(
        combinations(opt_idx, r) for r in range(len(opt_idx) + 1)
    ):
        # skip contains index of levels to skip
        # we go through every possible missing levels combinations
        parts = []
        for i, part in enumerate(schema["folders"]):
            if i in skip:
                continue
            if isinstance(part, str):
                if part.startswith("("):
                    part = part[1:-1]
                parts.append(_as_template(part))
            elif isinstance(part, dict):
                parts.append(part["text"])
            else:
                parts.append("_".join(map(_as_template, part)))
        raw_folders.append("/".join(parts))

    # # Inject conditions
    folders = raw_folders
    for conditions in schema["with"]:
        if "value" not in conditions:
            # This means that the facet must be set.
            # Not useful when parsing. Implicit with the facet in the pattern.
            continue

        # Ensure a list
        if isinstance(conditions["value"], str):
            value = [conditions["value"]]
        else:
            value = conditions["value"]

        patterns = []
        for patt in folders:
            for val in value:
                patterns.append(partial_format(patt, **{conditions["facet"]: val}))
        folders = patterns

    # # Inject parsing requirements (hardcoded :( )
    folders = [folder.replace("{variable}", "{variable:_}") for folder in folders]

    # # Filenames
    if "DATES" in schema["filename"]:
        if schema["filename"][-1] != "DATES":
            raise ValueError(
                "Reverse pattern generation is not supported for filenames with date bounds not at the end."
            )
        filename = "{?:_}_{DATES}"
    else:
        filename = "{?:_}"

    exts = exts or [".nc", ".zarr", ".zarr.zip"]

    patterns = [f"{fold}/{filename}{ext}" for fold, ext in product(folders, exts)]

    return patterns



# noqa: D100
import shutil
from pathlib import Path

import pytest

import xscen as xs
from xscen.testing import datablock_3d as _datablock_3d

notebooks = Path(__file__).parent.parent / "docs" / "notebooks"
SAMPLES_DIR = notebooks / "samples" / "tutorial"


@pytest.fixture(scope="session", autouse=True)
def cleanup_notebook_data_folder(request):
    """Cleanup a testing file once we are finished.

    This flag prevents remote data from being downloaded multiple times in the same pytest run.
    """

    def remove_data_folder():
        data = notebooks / "_data"
        if data.exists():
            shutil.rmtree(data)

    request.addfinalizer(remove_data_folder)


@pytest.fixture(scope="session")
def samplecat(request):
    """Generate a sample catalog with the tutorial netCDFs."""
    mark_skip = request.config.getoption("-m")
    if "not requires_netcdf" in mark_skip or not SAMPLES_DIR.exists():
        pytest.skip("Skipping tests that require netCDF files")
    elif not list(SAMPLES_DIR.rglob("*.nc")):
        pytest.skip("No netCDF files found in the tutorial samples folder")

    df = xs.parse_directory(
        directories=[SAMPLES_DIR],
        patterns=[
            "{activity}/{domain}/{institution}/{source}/{experiment}/{member}/{frequency}/{?:_}.nc"
        ],
        homogenous_info={
            "mip_era": "CMIP6",
            "type": "simulation",
            "processing_level": "raw",
        },
        read_from_file=["variable", "date_start", "date_end"],
        xr_open_kwargs={"engine": "h5netcdf"},
    )
    return xs.DataCatalog({"esmcat": xs.catalog.esm_col_data, "df": df})


@pytest.fixture
def datablock_3d():
    """Create a generic timeseries object based on pre-defined dictionaries of existing variables.

    See Also
    --------
    xscen.testing.datablock_3d
    """
    return _datablock_3d
